#!/usr/bin/env python
# -*- coding: iso8859-1 -*-
"""
   OneFinger - main module

   @license: GNU GPL, see COPYING for details
   @copyright: 2004, Maurizio Colucci
"""

import sys, os,  datetime, pickle, shutil, config, persistence, string, re, signal
import kdecore
import kdeui
import mimetypes
import kio
from sets import Set
from util import createDirIfNotExisting, find, addSlashIfNeeded, ext, first, forall, remove_ending_slash_if_not_root
from util import exists
from qt import QMessageBox, QPixmap, QIconSet, QObject, PYSIGNAL, SIGNAL, QVBox, QTimer, QStringList, QWidget
from qt import QHBox, QMainWindow, QApplication, QString, QTextEdit, QPushButton, QToolBar, QToolButton, QScrollView
from qt import QFrame, Qt, QSizePolicy,  QGrid, QLabel, QToolTip, QUriDrag, QTextDrag, qt_xrootwin, QFont, QPoint, QSize
from my_qt_widgets import FlashingPushButton, FlashingToolButton, MyPopupMenu, FlashingLabel, WrappingButton, MyListBox, MyLineEdit, MyComboBox, TextItem, TextSeparator, PopupMenuItem, FlashingTextEdit, text_item_button_info, EmptySeparator
from dcopexport import DCOPExObj


def _(x):
    return x





listBox = {}
sortBox = {}

filterBox = {}
filterButton = {}
filterMenu = {}
filterQuickText = {}
filterClearButton = {}

hBox = {}
icon = {}

last_clicked_program = None
files_whose_light_bulb_is_pressed = []
dragging_state = "no-drag"
programs_that_work_with_any_file = Set(['mv', 'cp', 'rm', 'konqueror', 'du', 'ln'])

message_you_can_now_select_a_file = _("You can now execute the command above, or click additional files or arguments.")
def point_to_string(p):
    if p is None:
        return None
    else:
        return p.x, p.y


class WindowInfo:
    def __init__(self, wid, name):
        self.wid = wid
        self.name = name
    def stringForPopupMenu(self):
        return self.name


class my_dcop_obj(DCOPExObj):
    def __init__(self):
        DCOPExObj.__init__(self, "activator")
        self.addMethod ("void activate()", self.activate)
    def activate(self):
        windowlist_window.move(QPoint(windowlist_x, windowlist_y))
        windowlist_window.show()
        kdecore.KWin.forceActiveWindow(windowlist_window.winId())
        # BEGIN hack to remove the highlighted item @@
        listBox['window'].lastMousePos = None
        listBox['window'].make_changes_visible() 
        # END

my_dcop_inst = my_dcop_obj()


class MyMainWindow(kdeui.KMainWindow):
    def queryClose(self):
        global recent_dirs
        recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
        return True # accept the closing


class MyListWindow(kdeui.KMainWindow):
    def moveEvent(self, ev):
        #print "moveEvent"
        global windowlist_x, windowlist_y
        windowlist_x = self.x()
        windowlist_y = self.y()

    def resizeEvent(self, ev):
        #print "resizeEvent"
        global windowlist_wt, windowlist_ht
        windowlist_wt = self.width()
        windowlist_ht = self.height()

    def closeEvent(self, ev):
        #print "closeEvent"
        ev.ignore()

last_cached_file = None
last_cached_file_is_file = None
last_cached_file_is_dir = None
def cached_is_file(x):
    global last_cached_file, last_cached_file_is_file, last_cached_file_is_dir
    if last_cached_file != x:
        last_cached_file = x
        last_cached_file_is_file = os.path.isfile(x)
        last_cached_file_is_dir = os.path.isdir(x)
    return last_cached_file_is_file
def cached_is_file_or_dir(x):
    global last_cached_file, last_cached_file_is_file, last_cached_file_is_dir
    if last_cached_file != x:
        last_cached_file = x
        last_cached_file_is_file = os.path.isfile(x)
        last_cached_file_is_dir = os.path.isdir(x)
    return last_cached_file_is_file or last_cached_file_is_dir
def cached_is_dir(x):
    global last_cached_file, last_cached_file_is_file, last_cached_file_is_dir
    if last_cached_file != x:
        last_cached_file = x
        last_cached_file_is_file = os.path.isfile(x)
        last_cached_file_is_dir = os.path.isdir(x)
    return last_cached_file_is_dir
        

def clearFileFiltersAndUpdate():
    clearAux('file')
def clearWindowFiltersAndUpdate():
    clearAux('window')
def clearArgumentFiltersAndUpdate():
    clearAux('argument')
def clearProgramFiltersAndUpdate():
    clearAux('program')

def clearAux(k):
    filterQuickText[k].clear()
    if k!= 'window':
        redraw_programs()
        redraw_files(centering = True)
        redraw_arguments()
    redraw_windows(remember_scroll = True)
    

class argument_info:
    def __init__(self, name, usage_count, last_usage_time):
        self.name = name
        self.usage_count = usage_count
        self.last_usage_time = last_usage_time

def file_to_icon(x):
    x2 = os.path.normpath(x)
    if cached_is_dir(x2):
        if '/' == x2:
            ic = QPixmap(config.imgDir + "file-manager.png")
        else:
            if os.path.ismount(x2):
                if is_substring_of('cd', x2) or is_substring_of('dvd', x2):
                    ic = QPixmap(config.imgDir + "cdrom.png")
                else:
                    ic = QPixmap(config.imgDir + "devices.png")
            else:
                if os.path.normpath( os.path.expanduser('~/')) == x2:
                    ic = QPixmap(config.imgDir + "home.png")
                else:
                    if x2 in important_directories:
                        ic = QPixmap(config.imgDir + "folder-favourite.png")
                    else:
                        ic = QPixmap(config.imgDir + "folder.png") 
    elif cached_is_file(x):
        ex = ext(x)
        if ex.endswith('~'):
            ex2 = ex[1:-1]
        else:
            ex2 = ex[1:]
        ic =  QPixmap(config.imgDir + typeToIcon.get(ex2, "generic.png"))
    else:
        ic = QPixmap()
    return ic

def redraw_arguments():
    def split_and_sort_args(pairs):
        k = 'argument'
        tx = sortBox[k].currentText()
        if tx == _("Name"):
            pairs.sort(lambda p1, p2: cmp(p1.name, p2.name))
            names = [x.name for x in pairs]
            return [(_('Sorted by name'), names)]
        elif tx == _('Recent usage'):
            pairs.sort(lambda p1, p2: -cmp(p1.last_usage_time, p2.last_usage_time))
            names = [x.name for x in pairs]
            return [  (_('Recently used'), names)]
        elif tx == _('Frequent usage'):
            pairs.sort(lambda p1, p2: -cmp(p1.usage_count, p2.usage_count))
            names = [x.name for x in pairs]
            return [  (_('Frequently used '), names)]
        else:
            raise "error"
    k = 'argument'
    listBox[k].clear()
    filt = str(filterQuickText[k].text())
    if filt != "":
        filter_string =  _("Containing \"") + filt + "\"; "
    else:
        filter_string = ""
    if show_files_in_arg_panel:
        not_a_file_string = ""
    else:
        not_a_file_string = "Not a file; "
    if last_clicked_program != None:
        pairs = arguments_of_program.get(last_clicked_program , [])
        pairs_filtered = [a for a in pairs if  is_substring_of(filt, a.name)]
        if not show_files_in_arg_panel:
            pairs_filtered2 = [a for a in pairs_filtered
                               if a.name[0] != '/']
        else:
            pairs_filtered2 = pairs_filtered
        l = split_and_sort_args(pairs_filtered2)
        trunc = 40 
        for j in range(0, len(l)):
            i = l[j]
            (sorting_string, args) = i
            if j != 0:
                listBox[k].items.append(EmptySeparator( listBox[k].listArea.width()))
            listBox[k].items.append(
                TextSeparator(_('Used with ' + last_clicked_program + "; ")
                              + not_a_file_string
                              + filter_string
                              + sorting_string
                              , listBox[k].listArea.width()))
            for x in args[:trunc]:
                if x in files_whose_light_bulb_is_pressed:
                    press = "pressed"
                else:
                    press = "released"

                if cached_is_dir(x):
                     bic = [ text_item_button_info( [ QPixmap(config.imgDir +  "enter-no-arrow.png")
                                                      , QPixmap(config.imgDir + "enter.png")]
                                                    , "enter directory"
                                                    , "not a togglebutton" )
                             , text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                                      , QPixmap(config.imgDir + "lit-bulb-small.png")]
                                                     , "narrow", press)]
                elif cached_is_file(x):
                     bic = [text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                              ,QPixmap(config.imgDir + "lit-bulb-small.png")]
                                             , "narrow", press)]
                else:
                    bic = []
                icon =  file_to_icon(x)

                listBox[k].items.append(TextItem(
                    text = x,
                    icon = icon,
                    buttons_info = bic,
                    isSelectable =   True #main_program_in_command(str(commandEntry.text())) != None,
                    ,parent_list = listBox[k]

                    , pressed = False
                    , data = x
                    , resolution_mult = resolution_multiplier
                    ))
            if len(args) > trunc:
                listBox[k].items.append(TextItem(
                    text = _("(The list was too long and has been truncated.)"),
                    icon = QPixmap(),
                    buttons_info = [],
                    isSelectable = False,
                    parent_list = listBox[k],
                    pressed = False,
                    data = ""
                    , resolution_mult = resolution_multiplier))
            if args == []:
                listBox[k].items.append(TextItem(text = "(no such argument exists)",
                                                 icon = QPixmap(),
                                                 buttons_info = [],
                                                 isSelectable = True,
                                                 parent_list = listBox[k],
                                                 pressed = False,
                                                 data = ""
                                                 , resolution_mult = resolution_multiplier))
    else:
        pairs = arguments_used_with_any_program
        pairs_filtered = [a for a in pairs if  is_substring_of(filt, a.name)]
        if not show_files_in_arg_panel:
            pairs_filtered2 = [a for a in pairs_filtered
                               if a.name[0] != '/']
        else:
            pairs_filtered2 = pairs_filtered

        l = split_and_sort_args(pairs_filtered2)
        trunc = 40
        for j in range(0, len(l)):
            i = l[j]
            (sorting_string, args) = i
            if j != 0:
                listBox[k].items.append(EmptySeparator( listBox[k].listArea.width()))
            listBox[k].items.append(
                    TextSeparator(_('Used with any program; ')
                                  + not_a_file_string
                                  + filter_string
                                  + sorting_string
                                  , listBox[k].listArea.width()))
            for x in args[:trunc]:
                if x in files_whose_light_bulb_is_pressed:
                    press = "pressed"
                else:
                    press = "released"

                if cached_is_dir(x):
                     bic = [ text_item_button_info( [ QPixmap(config.imgDir +  "enter-no-arrow.png")
                                                      , QPixmap(config.imgDir + "enter.png")]
                                                    , "enter directory"
                                                    , "not a togglebutton" )
                             , text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                                      , QPixmap(config.imgDir + "lit-bulb-small.png")]
                                                     , "narrow", press)]
                elif cached_is_file(x):
                     bic = [text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                              ,QPixmap(config.imgDir + "lit-bulb-small.png")]
                                             , "narrow", press)]
                else:
                    bic = []
                icon =  file_to_icon(x)
                listBox[k].items.append(TextItem(
                    text = x,
                    icon = icon,
                    buttons_info = bic,
                    isSelectable =  True #main_program_in_command(str(commandEntry.text())) != None,
                    , parent_list = listBox[k]
                    , pressed = False
                    , data = x
                    , resolution_mult = resolution_multiplier))
            if len(args) > trunc:
                listBox[k].items.append(TextItem(
                    text = _("(The list was too long and has been truncated.)"),
                    icon = QPixmap(),
                    buttons_info = [],
                    isSelectable = False,
                    parent_list = listBox[k],
                    pressed = False,
                    data = ""
                    , resolution_mult = resolution_multiplier))
            if args == []:
                listBox[k].items.append(TextItem(text = "(no such argument exists)",
                                                 icon = QPixmap(),
                                                 buttons_info = [],
                                                 isSelectable = True
                                                 , parent_list = listBox[k]
                                                 , pressed = False
                                                 , data = ""
                                                 , resolution_mult = resolution_multiplier))

    listBox[k].make_changes_visible()

def onFilterKeyTimer(widget):
    applyFilter()
    
def onEnterPressedOnFilter():
    applyFilter()
    
def applyFilter():
    global recent_filters
    redraw_programs()
    redraw_files()
    redraw_arguments()
    redraw_windows(remember_scroll = True)
    # BEGIN rebuild recent_filters[k] for all k
    for k in ['program', 'file', 'argument']:
        t = str(filterQuickText[k].text())
        if t:
            if t in recent_filters[k]:
                recent_filters[k].remove(t)
            recent_filters[k] = [t] + recent_filters[k]
        recent_filters[k] = (recent_filters[k])[:config.gMaxItemsInRecentLocMenu] # truncate
    # END
    rebuildFilterMenus()

def rebuildFilterMenus():
    for k in ['program', 'file', 'argument']:
        filterMenu[k].clear()
        l = recent_filters[k]
        l.reverse()
        for f in l:
            filterMenu[k].addItem(f, QPixmap(), lambda x:x)
        l.reverse()

def onRecentFilterClicked(menu, filter):
    print "onRecentFilterClicked"
    k = find(['argument', 'file', 'program'], lambda x: menu is filterMenu[x])
    filterQuickText[k].setText(filter)
    applyFilter()
         

              
def split_and_sort_files_or_progs(k, objs2):
    if k == 'file':
        last_usage = last_usage_time_of_dir
        usage_count = usage_count_of_dir
    elif k == 'program':
        last_usage = last_usage_time_of_program
        usage_count = usage_count_of_program
    tx = sortBox[k].currentText()
    if  tx == _("Name"):
        objs2.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        return [(_('Sorted by name'), objs2)]
    elif tx == _('Recent usage'):
        usedObjs = [ x for x in objs2 if last_usage.get(x, None) != None]
        unusedObjs = [x for x in objs2 if last_usage.get(x, None) == None]
        unusedObjs.sort(lambda o1, o2: cmp(o1.lower(),   o2.lower()))
        usedObjs.sort(lambda o1, o2: -cmp(last_usage[o1],last_usage[o2]))
        return [  (_('Recently used'), usedObjs) ,
                  (_('Never used'), unusedObjs)]
    elif tx == _("Frequent usage"):
        usedObjs = [x for x in objs2 if usage_count.get(x, 0) != 0]
        unusedObjs = [x for x in objs2 if usage_count.get(x, 0) == 0]
        unusedObjs.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        usedObjs.sort(lambda o1, o2: - cmp(usage_count[o1], usage_count[o2]))
        return [
                (_('Frequently used'), usedObjs),
                (_('Never used'), unusedObjs)]
    elif tx  == _("Date"):
        objs2.sort(lambda o1, o2: -cmp(os.path.getmtime(o1), os.path.getmtime(o2)))
        return [ (_('Most recently modified first'), objs2)]
    elif tx == _("Size"):
        withSize = [o for o in objs2 if os.path.isfile(o)]
        withoutSize = [ o for o in objs2 if not os.path.isfile(o)]
        withSize.sort(lambda o1, o2: -cmp(os.path.getsize(o1), os.path.getsize(o2)))
        withoutSize.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        return [ (_('Biggest first'), withSize),
                 (_('No size'), withoutSize)]
    else:
        raise "error: the current text is invalid: " + tx + " " + sortBox[k].currentText()

def onLocationButtonPressed():
    #command = str(commandEntry.text()) + cur_dir + " "
    t = make_string_bash_compatible(cur_dir)
    commandEntry.insert(t + "  ")
    commandEntry.setFocus()
    #commandEntry.moveCursor(QTextEdit.MoveEnd, False)
    commandEntry.startFlashing(3, True)


def pid(w):
    prop = [ kdecore.NET.WMPid, 0]
    inf = kdecore.NETWinInfo(   w
                                , qt_xrootwin()
                                , prop
                                , 2)
    return inf.pid()

def is_desktop_or_dock_or_utility_window(w):
    inf = kdecore.KWin.windowInfo(w, kdecore.NET.WMWindowType)
    t =  inf.windowType(kdecore.NET.NormalMask
                        | kdecore.NET.UtilityMask
                        | kdecore.NET.DesktopMask
                        | kdecore.NET.DockMask
                        | kdecore.NET.ToolbarMask)
    return t == kdecore.NET.Desktop or t==kdecore.NET.Dock or t==kdecore.NET.Utility


def redraw_windows(remember_scroll):
    if remember_scroll:
        scroll = listBox['window'].scrollingOffset
    listBox['window'].clear()
    def get_window_name(wid):
        inf = kdecore.KWin.windowInfo(wid)
        return inf.visibleName()

    tx = sortBox['window'].currentText()
    if tx == _("Recent usage"):
        windows = windows_sorted_by_usage
    else:
        windows = windows_sorted_by_creation_date
    interesting_windows = [ x for x in windows
                            if x != windowlist_window.winId()
                            if not is_desktop_or_dock_or_utility_window(x)]
    for w in interesting_windows:
        name = get_window_name(w)
        filt = str(filterQuickText['window'].text())
        if is_substring_of(filt, name) and w != windowlist_window.winId():
            listBox['window'].items.append(
                TextItem(text = name
                         , icon = kdecore.KWin.icon(w, 26, 26, True)
                         , buttons_info = [] 
                         , isSelectable = True
                         , parent_list =   listBox['window']
                         , pressed = False
                         , data = w
                         , resolution_mult = resolution_multiplier))
    if remember_scroll:
        listBox['window'].scrollingOffset = scroll
#     if   main_program_in_command(str(commandEntry.text())) != None:
#         listBox['window'].enable_all_items(True)
#     else:
#         listBox['window'].enable_all_items(False)
    #print "redraw_windows: calling make_changes_visible"
    listBox['window'].make_changes_visible()

def redraw_files(centering = False):
    # BUG: @@ cur_dir might stop being an existing directory anywhere
    # during the execution of this function. This is because the user
    # could be running a "rm -rf" command in background.
    k = 'file'
    if centering:
        offs = listBox['file'].scrollingOffset
    locationButton.setText(cur_dir)
    locationButton.data = cur_dir
    set_current_dir_button.setIconSet(QIconSet(QPixmap(file_to_icon(cur_dir))))
    #locationButton.setEnabled( main_program_in_command(str(commandEntry.text())) != None)
    set_current_dir_button.setOn(cur_dir == os.getcwd())
    narrow_dir_button.setOn(cur_dir in files_whose_light_bulb_is_pressed)
    listBox['file'].clear()
    cur_dir_slash = addSlashIfNeeded(cur_dir)
    files_without_path = [x for x in os.listdir(cur_dir)]
    files_with_path = [ cur_dir_slash + x for x in files_without_path]
    dirs = [o for o in files_with_path if os.path.isdir(o)]
    ordinary_files = [o for o in files_with_path if os.path.isfile(o)]
    executable_files = [o for o in ordinary_files if
                        os.access(o, os.X_OK)]
    filt = str(filterQuickText['file'].text())
    interesting_files = [x for x in dirs + ordinary_files if is_substring_of(filt, x)]
    l = split_and_sort_files_or_progs('file', interesting_files)
    previous_empty = True
    for i in l:
        (title, files) = i
        if files != []:
            if not previous_empty:
                listBox[k].items.append(EmptySeparator( listBox[k].listArea.width()))
            previous_empty = False
            listBox['file'].items.append(TextSeparator(title, listBox[k].listArea.width()))
        else:
            previous_empty = True
        for x in files:
            #BEGIN make icons (ic =icon, bic = button icon)
            if x in files_whose_light_bulb_is_pressed:
                press = "pressed"
            else:
                press = "released"
            ic = file_to_icon(x)
            if x in dirs:
                #ic = QPixmap(config.imgDir +  "folder.png")
                bic = [ text_item_button_info( [ QPixmap(config.imgDir +  "enter-no-arrow.png")
                                                 , QPixmap(config.imgDir + "enter.png")]
                                               , "enter directory"
                                               , "not a togglebutton" )
                        , text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                                 , QPixmap(config.imgDir + "lit-bulb-small.png")]
                                                , "narrow", press)]
            else:
                #ex = ext(x)
                #if ex.endswith('~'):
                #    ex2 = ex[1:-1]
                #else:
                #    ex2 = ex[1:]
                #ic =  QPixmap(config.imgDir + typeToIcon.get(ex2, "generic.png"))
                
                bic = [text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                              ,QPixmap(config.imgDir + "lit-bulb-small.png")]
                                             , "narrow", press)]
            # END
            listBox['file'].items.append(TextItem(
                text = os.path.basename(x),
                icon = ic,
                buttons_info = bic,
                isSelectable = True # ( main_program_in_command(str(commandEntry.text())) != None or x in executable_files)
                , parent_list = listBox['file']
                , pressed = False
                , data = x
                , resolution_mult = resolution_multiplier))
    # BEGIN scroll the list to restore the previous scrolling position
    if centering:
        listBox['file'].scrollingOffset = offs
#        (text, offset) = centering
#        listBox['file'].centerItem(text, offset)
    #END
    listBox['file'].make_changes_visible()

def is_substring_of(s, st):
    return st.lower().find(s.lower()) != -1


program_icons = {

    "xpdf": "xpdf.png"
    , "gnome-terminal": "gnome-terminal.png"
    , "konsole": "konsole.png"
    , "aterm": "konsole.png"
    , "none": "italian-flag.png"
    , "kfind": "kfind.png"
    , "kedit": "kwrite.png"
    , "kate": "kwrite.png"
    , "kbear": "kbear.png"
    , "gedit": "gedit.png"
    , "kwrite": "kwrite.png"
    , "krename": "name.png"
    , "kmix": "kmix.png"
    , "aumix": "aumix.png"
    , "ksnapshot": "ksnapshot.png"
    , "gnome-screenshot": "gnome-screenshot.png"
    , "kcontrol": "kcontrol.png"
    , "kparted": "devices.png"
    , "qtparted": "devices.png"
    , "diskdrake": "diskdrake.png"
    , "kdf": "kdiskfree.png"
    , "pan": "pan.png"
    , "knews": "news.png"
    , "dvdrip": "cdrom.png"
    , "kgpg": "kgpg.png"
    , "ark": "ark.png"
    , "kuickshow": "kuickshow.png"
    , "eog": "eog.png"
    , "gthumb": "gthumb.png"
    , "korganizer": "korganizer.png"
    , "sylpheed": "sylpheed.png"
    , "abiword": "abiword.png"
    , "nautilus": "nautilus.png"
    , "gftp": "gftp.png"
    , "kpdf": "generic.png"
    , "kghostview": "kghostview.png"
    , "kdvi": "kdvi.png"
    , "gpdf":"gpdf.png"
    , "gimp": "gimp.png"
    , "k3b": "k3b.png"
    , "emacs": "emacs.png"
    , "cervisia": "cervisia.png"
    , "mplayer": "mplayer.png"
    , "oowriter": "ooowriter.png"
    , "texmacs": "texmacs.png"
    , "konqueror": "konqueror.png"
    , "firefox": "firefox.png"
    , "mozilla-firefox": "firefox.png"
    , "mozilla-thunderbird": "mozicon16.xpm"
    , "thunderbird": "mozicon16.xpm"
    , "ksysguard": "kdesystemguard.png"
    , "mozilla": "mozilla.png"
    , "opera": "opera.png"
    , "evolution": "evolution.png"
    , "kmail": "kmail.png"
    , "lyx": "lyx.png"
    , "kopete": "kopete.png"
    , "knode": "knode.png"
    , "amule": "amule.png"
    , "synaptic": "kynaptic.png"
    , "kynaptic": "kynaptic.png"
    , "gmplayer": "mplayer.png"
    , "umbrello": "umbrello.png"
    , "xine": "xine.png"
    , "xchat": "xchat.png"
    , "xcalc": "xcalc.png"
    , "file-roller": "fileroller.png"
    , "totem": "totem.png"
    , "xmms": "xmms.png"
    , "kaffeine": "kaffeine.png"
    , "gxine": "xine.png"
    }

def types_selected_for_prog_narrowing():
    s = Set()
    for f in files_whose_light_bulb_is_pressed:
        if os.path.isdir(f):
            s.add("<directory>")
        else:
            (type, encod) = mimetypes.guess_type(f)
            if type != None:
                s.add(type)
            else:
                ex = ext(f)
                s.add(ex)

    return s

def redraw_programs():
    def set_to_string(s):
        l = []
        for x in s:
            l.append( x)
        st = "\""
        for x in range(0, len(l)):
            if l[x] == "":
                l[x] = "<no type>"
            st += l[x]
            if x<len(l)-1:
                st += ", "
        return st + "\""
    k = 'program'
    listBox[k].clear()
    filt = str(filterQuickText[k].text())
    filtered_progs = [p for p in programs
                      if is_substring_of(filt, p)
                      if ( forall(types_selected_for_prog_narrowing(),
                                  lambda e: ( e in types_used_with_program.get(p, Set())
                                              or
                                              e in types_usable_with_program.get(p, Set())))
                           or
                           p in programs_that_work_with_any_file)]
    l = split_and_sort_files_or_progs(k, filtered_progs)
    trunc = 100
    previous_empty = True
    for i in l:
        (sorting_string, progs) = i
        if progs != [] or len(types_selected_for_prog_narrowing())> 0:
            if filt != "":
                filter_string =  _("Containing \"") + filt + "\"; "
            else:
                filter_string = ""
            if len(types_selected_for_prog_narrowing())>0:
                narrowing_string = ( _("Can manage ")
                                     + set_to_string(types_selected_for_prog_narrowing())
                                     + "; ")
            else:
                narrowing_string = ""
            if not previous_empty:
                listBox[k].items.append(EmptySeparator( listBox[k].listArea.width()))
            previous_empty = False
            listBox[k].items.append(TextSeparator(narrowing_string
                                                  +filter_string 
                                                  + sorting_string
                                                  , listBox[k].listArea.width()))
        else:
            previous_empty = True
        for x in progs[:trunc]:
            ic =  QPixmap(config.imgDir + program_icons.get(x, "generic.png"))
            
            listBox[k].items.append(TextItem(
                text = x,
                icon = ic,
                buttons_info = [text_item_button_info([QPixmap(config.imgDir + "glasses.png")
                                                       , QPixmap(config.imgDir + "glasses.png")]
                                                      , "show arguments"
                                                      , "not a togglebutton"
                                                      )],
                isSelectable = True,
                parent_list = listBox[k]
                , pressed = False
                , data = x
                , resolution_mult = resolution_multiplier))
        if len(progs) > trunc:
            listBox[k].items.append(TextItem(
                text = _("(The list was too long and has been truncated. Use the filter area (below) to find the program you need. You should already know what program you need.)"),
                icon = QPixmap(),
                buttons_info = [],
                isSelectable = False,
                parent_list = listBox[k],
                pressed = False,
                data = ""
                , resolution_mult = resolution_multiplier))
            
    listBox[k].make_changes_visible()

typeToIcon = {
    'txt': 'text-file.png',
    'ics': 'schedule.png',
    'pdf': 'pdf.png',
    'dvi': 'kdvi.png',
    'ps':  'kdvi.png',
    'doc': 'word-26x26.png',
    'rpm': 'rpm.png',
    'mp3': 'audio.png',
    'flac': 'audio.png',
    'ogg': 'audio.png',
    'wav': 'audio.png',
    'mpc': 'audio.png',
    'flac': 'audio.png',
    'ogm': 'video.png',
    'avi': 'video.png',
    'mkv': 'video.png',
    'mpg': 'video.png',
    'mpeg': 'video.png',
    'sxw': 'ooowriter.png',
    'wma': 'audio.png',
    'wmv': 'video.png',
    'spec': 'rpm.png',
    'asf': 'video.png',
    'zip': 'archive.png',
    'rar': 'archive.png',
    'tgz': 'archive.png',
    'gz':  'archive.png',
    'bz2': 'archive.png',
    'tar': 'archive.png',
    'arj': 'archive.png',
    'ace': 'archive.png',
    'htm': 'internet.png',
    'html': 'html.png',
    'png': 'picture.png',
    'gif': 'picture.png',
    'jpg': 'picture.png',
    'ppm': 'picture.png',
    'bmp': 'picture.png',
    'xpm': 'picture.png',
    'blend': 'picture.png',
    'py':'program-listing.png',
    'pl':'program-listing.png',
    'm':'program-listing.png',
    'cc':'program-listing.png',
    'hpp':'program-listing.png',
    'h':'program-listing.png',
    'c':'program-listing.png',
    'cpp':'program-listing.png',
    'c++':'program-listing.png',
    'pas':'program-listing.png',
    'ml':'program-listing.png',
    'bas':'program-listing.png'
}



def add_cur_dir_to_recent_dirs(recent_dirs):
    r = find(recent_dirs, lambda x: x== cur_dir)
    if not r is None:
        recent_dirs.remove(cur_dir)
    recent_dirs2 = [cur_dir] + recent_dirs
    rebuild_recent_locations_menu(recent_dirs2)
    return recent_dirs2[:config.gMaxItemsInRecentLocMenu]
  

def rebuild_parent_dirs_menu():
    
    def get_parent_dirs(d):
         l = []
         cur_dir2 = remove_ending_slash_if_not_root(cur_dir)
         (a, b) = os.path.split(cur_dir2)
         while b != '':
             l.append(a)
             (a, b) = os.path.split(a)

         return l
    parentLocPopupMenu.clear()
    parent_dirs = get_parent_dirs(cur_dir)
    for l in parent_dirs:
        l2 = remove_ending_slash_if_not_root(l)
        ic = file_to_icon(l2)
        parentLocPopupMenu.addItem(l2, ic, lambda x:x)
    

def rebuild_recent_locations_menu(recent_dirs):
    recentLocPopupMenu.clear()
    for l in recent_dirs:
        l2 = remove_ending_slash_if_not_root(l)
        ic = file_to_icon(l2)
        recentLocPopupMenu.addItem(l2, ic, lambda x:x)

def rebuild_important_directories_menu():
    global important_directories
    important_directories.sort(lambda d1, d2: -cmp(usage_count_of_dir.get(d1,0)
                                                   , usage_count_of_dir.get(d2,0)))
    favouriteLocPopupMenu.clear()
    for l in important_directories:
        l2 = remove_ending_slash_if_not_root(l)
        ic = file_to_icon(l2)
        favouriteLocPopupMenu.addItem(l2, ic, lambda x:x)
        
# BEGIN Qt only
#applic = QApplication(sys.argv)
# END

# BEGIN KDE only
# this must be before the creation of MyMainWindow
kdecore.KCmdLineArgs.init(sys.argv, "OneFinger", "Desktop environment", config.version)
applic = kdecore.KUniqueApplication(True, True, True)
# END


# BEGIN detect resolution
real_resolution =  kdecore.KApplication.desktop().width()

resolution_multiplier = float(real_resolution/1024.0)

# END

createDirIfNotExisting(config.dataDir)

def rescan_programs():
    def read_mimetypes_and_rebuild_types_usable_with_program():
        def extract_program_name(cmd):
            tokens = string.split(str(cmd))
            return os.path.basename(tokens[0])
        global types_usable_with_program
        # BEGIN hardcoded knowledge about programs
        types_usable_with_program = { 'kuickshow': Set(['<directory>'])
                                      , 'konqueror': Set(['<directory>'])
                                      , 'kfind': Set(['<directory>'])
                                      , 'konsole': Set(['<directory>'])
                                      , 'kdirstat': Set(['<directory>'])
                                      , 'filelight': Set(['<directory>'])

                                      }
        # END
        services = kio.KService.allServices()
        programs = [x for x in services if x.type_() == "Application"
                    if len(str(x.exec_())) > 0]
        for p in programs:
            #         print s.name()
            #         print "generic name = " + str(s.genericName())
            #         print "exec = " + str(s.exec_())
            #         print "name guessed = "+ str(extract_program_name(s.exec_()))
            #         print s.comment()
            #         print "icon = " + str(s.icon())
            #print "path = " + str(s.path())
            p2 = extract_program_name(p.exec_())
            if types_usable_with_program.get(p2,None) == None:
                types_usable_with_program[p2] = Set()
            for s in p.serviceTypes():
                types_usable_with_program[p2].add(str(s))

    global programs
    programs = Set()
    pathlist = os.environ['PATH'].split(os.pathsep)
    for path in pathlist:
        # find prgs really there and allowed to execute
        if os.path.isdir(path): # covers the case where the user has an unexisting path in $PATH
            for p in os.listdir(path):
                if os.access(os.path.join(path,p), os.X_OK):
                    programs.add(p)
    read_mimetypes_and_rebuild_types_usable_with_program()

rescan_programs()


# BEGIN build default arguments for programs
default_arguments = [ ("mplayer", [ "-vo", "sdl", "xv", "--enqueue", "-sub"])
                      ,("gmplayer", [ "-vo", "sdl", "xv", "--enqueue", "-sub"])
                      ,("totem", [ "--enqueue"])
                      ,("konsole", [ "--workdir"])
                      ,("xmms", [ "--enqueue"])
                      ,("k3b", [ "--emovixdvd", "--copycd", "--clonecd", "--erasecd", "--isoimage", "--binimage",
                                 "--formatdvd", "--datacd", "--audiocd", "--videocd", "--mixedcd", "--datadvd" ])
                      ,("xine", [ "--enqueue"])
                      ,("ark", [ "--extract", "--extract-to", "--add", "--add-to", "--guess-name"])
                      ,("file-roller", [ "--extract", "--extract-to=", "--add", "--add-to="])
                      ,("cp", [ "-r"])
                      ,("scp", [ "-r"])
                      ,("apt-cache", [ "show", "search"])
                      ,("dpkg", [ "-c","-i", "--install",  "--contents"])
                      ,("rpm", [ "-Uvh","-i", "-qpl", "--install"])
                      ,("tar", [ "-xzvf","-czvf", "-xjvf", "-cjvf"])
                      ,("mv", [ "-r"])
                      ,("du", [ "-s", "-h", "-sh", "--human-readable", "--summarize"])
                      ,("rm", [ "-r", "-f", "-rf"])
                      ,("btdownloadcurses", [ "--url"])
                      ,("chmod", [ "-R", "root:root", "*"])
                      ,("chown", [ "-R", "root:root", "*"])
                      ,("find", [ "-iname", "-exec", "{}", "-type d", "-type f" ])
                      ,("btdownloadgui", [ "--url"])
                      ,("df", [ "--human-readable"])
                      ,("ls", [ "--color", "-a", "-l", "-la"])
                      ,("emacs", [ "-nw"])
                      ,("echo", [ "$PATH"])
                      ,("apt", [ "search", "install", "--no-checksig"])
                      ,("apt-get", [ "update", "upgrade", "clean", "autoclean", "install"])
                      ,("update-alternatives", [ "--config", "x-window-manager", "x-terminal-emulator"
                                                 ,"x-www-browser", "x-session-manager", "editor"
                                                 , "emacs", "c++", "cc", "cli", "ctags", "editor", "dia"])
                      ]
default_arguments2 = {}
for pair in default_arguments:
    (prog_name, arg_list) = pair
    l2 = [argument_info(x, 1, datetime.datetime.now()) for x in arg_list]
    default_arguments2[prog_name] = l2

# END

cached_command = None
cached_tokens = []
def split_spaces_but_not_backslash_space(cmd):
    global cached_command
    global cached_tokens
    if cmp(cmd, cached_command)==0:
        #print "cache hit"
        return cached_tokens
    else:
        #print "cache miss"
        cached_command = cmd
        #print "invoking parser on " + cmd
        file =   os.popen(u'python ' + config.parser_file_dir + u'/parser.py ' + cmd, 'r')
        lines = file.readlines()
        file.close()
        cached_tokens = [string.strip(x) for x in lines[1:]]
        #print "result of parsing: "
        #print cached_tokens
        return cached_tokens


        


def parse_arguments_used_with_any_program_from_bash_profile():
#     file = os.open('~/.bash_profile', 'r')
#     lines = file.readlines()
#     file.close()
#     print lines
    return []


# BEGIN read database
windowlist_x_read_from_db = persistence.loadVariableFromFile(config.dataDir + 'windowlist_x'
                                                   , default = None)
windowlist_y_read_from_db = persistence.loadVariableFromFile(config.dataDir + 'windowlist_y'
                                                   , default = None)
windowlist_wt_read_from_db = persistence.loadVariableFromFile(config.dataDir + 'windowlist_wt'
                                                   , default = None)
windowlist_ht_read_from_db = persistence.loadVariableFromFile(config.dataDir + 'windowlist_ht'
                                                   , default = None)
help_bar_visible =persistence.loadVariableFromFile(config.dataDir + 'help_bar_visible'
                                                   , default = True)
arguments_used_with_any_program =persistence.loadVariableFromFile(
    config.dataDir + 'arguments_used_with_any_program'
    , default = parse_arguments_used_with_any_program_from_bash_profile())
types_used_with_program = persistence.loadVariableFromFile(
    config.dataDir + 'types_used_with_program'
    , default = {})
recent_dirs = persistence.loadVariableFromFile(config.dataDir + 'recent_dirs', default = [])
program_runs_in_terminal = persistence.loadVariableFromFile(config.dataDir + 'program_runs_in_terminal', default = {})
cur_dir = persistence.loadVariableFromFile(config.dataDir + 'cur_dir', default = '/')
arguments_of_program = persistence.loadVariableFromFile(
    config.dataDir + 'arguments_of_program',
    default = default_arguments2)
important_directories = persistence.loadVariableFromFile(config.dataDir
                                                         + 'important_directories'
                                                         , default = ['/'])
last_usage_time_of_dir = persistence.loadVariableFromFile(config.dataDir + 'last_usage_time_of_dir', default = {})
last_usage_time_of_program = persistence.loadVariableFromFile(
    config.dataDir + 'last_usage_time_of_program', default = {})
recent_filters = persistence.loadVariableFromFile(config.dataDir + 'recent_filters',
                                                  default = {'file': [], 'argument':[], 'program':[], 'window':[]})
sorting_criterion_for_dir = persistence.loadVariableFromFile(config.dataDir + 'sorting_criterion_for_dir', default = {})
sorting_criterion_for_arguments = persistence.loadVariableFromFile(
    config.dataDir + 'sorting_criterion_for_arguments'
    , default = "Recent usage")
sorting_criterion_for_programs = persistence.loadVariableFromFile(
    config.dataDir + 'sorting_criterion_for_programs'
    , default = "Frequent usage")

usage_count_of_dir = persistence.loadVariableFromFile(config.dataDir + 'usage_count_of_dir', default = {})
usage_count_of_program = persistence.loadVariableFromFile(config.dataDir + 'usage_count_of_program', default = {})
# END


mainWindow = MyMainWindow()
#mainWindow.setCaption("OneFinger - the graphical shell")
#mainWindow.setIcon(QPixmap(config.imgDir + "onefinger.png")) # ignored for a kde bug
applic.setMainWidget(mainWindow)

spacing = int (3.0 * resolution_multiplier)

windowlist_window = MyListWindow(mainWindow)
windowlist_window.setCaption("Window list")

windowlist_vbox = QVBox(windowlist_window)
windowlist_vbox.show()
windowlist_vbox.setSpacing(spacing)
windowlist_window.setCentralWidget(windowlist_vbox)


windowlist_help_hbox = QHBox(windowlist_vbox)
windowlist_help_hbox.setMargin(spacing * 2.0)
windowlist_help_hbox.setPaletteBackgroundColor(windowlist_help_hbox.palette().active().mid())
windowlist_help_hbox.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
windowlist_help_hbox.setSpacing(spacing)
windowlist_help_hbox.show()
windowlist_default_message = _("Click an item below to activate the corresponding window.")
windowlist_help_label = FlashingLabel( windowlist_default_message
                                       , windowlist_help_hbox)
windowlist_help_label.setAlignment(Qt.AlignCenter |Qt.AlignVCenter)
windowlist_help_label.startFlashing(config.gFlashingTimeForHelpMessage
                                    , resetOriginalColorsWhenFinished = True)


cancel_drag_button=QPushButton(QIconSet(QPixmap(config.imgDir + "delete.png"))
                               , _("Cancel drag")
                               , windowlist_help_hbox)
cancel_drag_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
cancel_drag_button.hide()

def on_cancel_drag():
    global dragging_state
    dragging_state = 'no-drag'
    cancel_drag_button.hide()
    windowlist_help_label.setText(windowlist_default_message)
    windowlist_help_label.startFlashing(config.gFlashingTimeForHelpMessage
                                        , resetOriginalColorsWhenFinished = True)

QObject.connect(cancel_drag_button, SIGNAL("clicked()"), on_cancel_drag)
#listBox['window'] =  MyListBox(frameVisible = False
#                               , parent = windowlist_vbox
#                               , resolution_mult = resolution_multiplier)



mainVBox = QVBox(mainWindow)
mainVBox.show()
mainVBox.setMargin(spacing)
mainVBox.setSpacing(spacing)
mainWindow.setCentralWidget(mainVBox)


directories_toolbar = QToolBar(mainWindow) # Ktoolbar is buggy: it
                                           # shows all its buttons
                                           # when shown, and it
                                           # doesn't stretch its
                                           # buttons horizontally!
directories_toolbar.show()
#directories_toolbar.setMargin(spacing)

misc_toolbar = QToolBar(mainWindow)
misc_toolbar.setHorizontallyStretchable(True)
misc_toolbar.show()


height_of_toolbar_buttons = int (50.0 * resolution_multiplier)
height_of_execute_buttons = int (50.0 * resolution_multiplier)

dir_toolbar_hbox = QHBox(directories_toolbar)
dir_toolbar_hbox.setSpacing(spacing)
#dir_toolbar_hbox.setMargin(spacing)

bookmarksButton = FlashingPushButton(QIconSet(QPixmap(config.imgDir +"folder-favourite.png")),
                 _("Important directories"),
                 dir_toolbar_hbox)
bookmarksButton.setMinimumHeight(height_of_toolbar_buttons)
QToolTip.add(bookmarksButton,

 _("Use this to change directory. The contents of the\ndirectory will appear in the panel below and to the left. \nIn order to add/remove a directory to/from this list\nyou can use the toolbar buttons to the right."))


favouriteLocPopupMenu = MyPopupMenu(bookmarksButton)



def onLocationSelectedFromPopupMenu(menu, new_dir):
    global recent_dirs, usage_count_of_dir, cur_dir
    recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
    cur_dir = new_dir
    filterQuickText['file'].clear()
    usage_count_of_dir[new_dir] = usage_count_of_dir.get(new_dir, 0) +1
    last_usage_time_of_dir[new_dir] = datetime.datetime.now()
    rebuild_recent_locations_menu(recent_dirs)
    rebuild_parent_dirs_menu()
    cr = _(sorting_criterion_for_dir.get(new_dir, "Recent usage")) # sic!
    sortBox['file'].setCurrentText(cr)
    redraw_files()
    listBox['file'].flash_all_items(config.gFlashingTimeForList)
    recent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                    resetOriginalColorsWhenFinished = True)
    parent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                    resetOriginalColorsWhenFinished = True)

    rebuild_important_directories_menu()

QObject.connect(favouriteLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)
bookmarksButton.setPopup(favouriteLocPopupMenu)
bookmarksButton.show()

#
recent_dirsButton = FlashingPushButton(QIconSet(QPixmap(config.imgDir +"folder-recent.png")),
                                       _("Recent directories"), dir_toolbar_hbox)
recent_dirsButton.setMinimumHeight(height_of_toolbar_buttons)
recent_dirsButton.show()
QToolTip.add(recent_dirsButton,  _("Use this to change directory. The contents of the\ndirectory will appear in the panel below and to the left."))

recentLocPopupMenu = MyPopupMenu(recent_dirsButton)
recent_dirsButton.setPopup(recentLocPopupMenu)
QObject.connect(recentLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)

#

parent_dirsButton = FlashingPushButton(QIconSet(QPixmap(config.imgDir +"folder-parent.png")),
                                   _(""), dir_toolbar_hbox)
parent_dirsButton.setMinimumHeight(height_of_toolbar_buttons)
parent_dirsButton.show()
QToolTip.add(parent_dirsButton,  _("Go to a parent directory."))

parentLocPopupMenu = MyPopupMenu(parent_dirsButton)
parent_dirsButton.setPopup(parentLocPopupMenu)
QObject.connect(parentLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)

#

class DragTargetButtonForUnicodeUri(QToolButton):
    def __init__(self, iconset, text, callback, parent ):
        QToolButton.__init__(self, iconset, text, "group text - unused", callback, parent)
        self.setAcceptDrops(True)
    def dragEnterEvent(self, event):
        event.accept(  QUriDrag.canDecode(event) or QTextDrag.canDecode(event))
    def dropEvent(self, event):
        print "dropEvent"
        l = QStringList()
        s = QString()
        if  QUriDrag.decodeToUnicodeUris(event, l):
            print "accept. l = " + str(l)
            event.acceptAction()
            for i in l:
                commandEntry.insert("\"" + str(i) + "\"  ")
        elif  QTextDrag.decode(event, s):
            print "accept. l = " + str(s)
            event.acceptAction()
            commandEntry.insert("\"" + str(s) + "\"")
        else:
            event.acceptAction(False)


def on_drag_button_pressed():
    QMessageBox.information(mainVBox, _("Drag and drop"),

                            _("This button is only here for dragging files onto it. Anything you\ndrag on it will be copied in the command area, in an appropriate\nformat (takes care of quoting, etc.)."  ))

drop_target = DragTargetButtonForUnicodeUri(QIconSet(QPixmap(config.imgDir + "stock_drag-mode.png"))
                                            , _("")
                                            , on_drag_button_pressed
                                            , misc_toolbar)
drop_target.setMinimumHeight(height_of_toolbar_buttons)
drop_target.show()
QToolTip.add(drop_target,

_("You can drop files, directories, http and ftp locations onto\nthis button. They will be copied into the command area in an\nappropriate format, so you can use them as part of a command."))


#QObject.connect(drop_target, SIGNAL("clicked()"), on_drag_button_pressed)


#
def on_refresh_files_pressed():
    redraw_files()
    listBox['file'].flash_all_items(config.gFlashingTimeForList)
refresh_files_button = QToolButton(QIconSet(QPixmap(config.imgDir + "refresh.png"))
                                   , _("")
                                   , "group text - unused"
                                   , on_refresh_files_pressed
                                   , misc_toolbar)
refresh_files_button.setMinimumHeight(height_of_toolbar_buttons)
refresh_files_button.show()
QToolTip.add(refresh_files_button,
             _("Refresh files. Click here to update the file list (useful in case some external program modified the files on the disk)."))

#

def save_database():
    # BEGIN write database
    print "Saving database..."
    persistence.dumpToFile(targetFile = config.dataDir + 'windowlist_x'
                           , var = windowlist_x)
    persistence.dumpToFile(targetFile = config.dataDir + 'windowlist_y'
                           , var = windowlist_y)
    persistence.dumpToFile(targetFile = config.dataDir + 'windowlist_wt'
                           , var = windowlist_wt)
    persistence.dumpToFile(targetFile = config.dataDir + 'windowlist_ht'
                           , var = windowlist_ht)
    persistence.dumpToFile(targetFile = config.dataDir + 'recent_dirs'
                           , var = recent_dirs)
    persistence.dumpToFile(targetFile = config.dataDir + 'program_runs_in_terminal'
                           , var = program_runs_in_terminal)
    persistence.dumpToFile(targetFile = config.dataDir + 'cur_dir', var = cur_dir)
    persistence.dumpToFile(targetFile = config.dataDir + 'important_directories'
                           , var = important_directories)
    persistence.dumpToFile(targetFile = config.dataDir + 'usage_count_of_dir'
                           , var = usage_count_of_dir)
    persistence.dumpToFile(targetFile = config.dataDir + 'types_used_with_program',
                           var = types_used_with_program)
    persistence.dumpToFile(targetFile = config.dataDir + 'usage_count_of_program'
                           , var = usage_count_of_program)
    persistence.dumpToFile(targetFile = config.dataDir + 'arguments_of_program'
                           , var = arguments_of_program)
    persistence.dumpToFile(targetFile = config.dataDir + 'recent_filters'
                           , var = recent_filters)
    persistence.dumpToFile(targetFile = config.dataDir + 'last_usage_time_of_dir'
                           , var = last_usage_time_of_dir)
    persistence.dumpToFile(targetFile = config.dataDir
                           + 'last_usage_time_of_program'
                           , var = last_usage_time_of_program)
    persistence.dumpToFile(targetFile = config.dataDir
                           + 'sorting_criterion_for_dir'
                           , var = sorting_criterion_for_dir)
    persistence.dumpToFile(targetFile = config.dataDir
                           + 'help_bar_visible', var = help_bar_visible)
    persistence.dumpToFile(targetFile = config.dataDir
                           + 'sorting_criterion_for_arguments'
                           , var = sorting_criterion_for_arguments)
    persistence.dumpToFile(targetFile = config.dataDir + 'sorting_criterion_for_programs'
                           , var = sorting_criterion_for_programs)
    persistence.dumpToFile(targetFile = config.dataDir + 'arguments_used_with_any_program'
                           , var = arguments_used_with_any_program)

    print "Database saving completed."
    # END



save_db_button = QToolButton(QIconSet(QPixmap(config.imgDir + "floppy.png"))
                                   , _("")
                                   , "group text - unused"
                                   , save_database
                                   , misc_toolbar)
save_db_button.setMinimumHeight(height_of_toolbar_buttons)
save_db_button.show()
QToolTip.add(save_db_button,
             _("Save the database now so that, if your system hangs, OneFinger will remember the most recent\nactions nonetheless."))

    
#QObject.connect(refresh_files_button, SIGNAL("clicked()"), on_refresh_files_pressed)

def on_rescan_programs_pressed():
    rescan_programs()
    redraw_programs()
    #program_label.startFlashing(config.gFlashingTimeForButtons
    #                            , resetOriginalColorsWhenFinished = True)
    listBox['program'].flash_all_items(config.gFlashingTimeForList)
rescan_programs_buttons = QToolButton(QIconSet(QPixmap(config.imgDir + "person.png"))
                                      , _("")
                                      , "group text - unused"
                                      , on_rescan_programs_pressed
                                      , misc_toolbar)
rescan_programs_buttons.setMinimumHeight(height_of_toolbar_buttons)
rescan_programs_buttons.show()
QToolTip.add(rescan_programs_buttons,
             _("Rescan programs. Click here when you have installed or removed a program, or if the $PATH has changed. \nIt rescans the $PATH and the MIME associations."))

#QObject.connect(rescan_programs_buttons, SIGNAL("clicked()"), on_rescan_programs_pressed)


#
def on_add_to_important_dir_clicked():
    print "here"
    global important_directories
    if not (cur_dir in important_directories):
        important_directories.append(cur_dir)
        redraw_files() # update icon (folder with heart)
        set_current_dir_button.startFlashing(config.gFlashingTimeForButtons * 3,
                                             resetOriginalColorsWhenFinished = True)
    rebuild_important_directories_menu()
    bookmarksButton.startFlashing(config.gFlashingTimeForButtons,
                                  resetOriginalColorsWhenFinished = True)

add_to_important_dirs_button = QToolButton(QIconSet(QPixmap(config.imgDir + "gnome-favorites.png"))
                                           , _("")
                                           , "unused"
                                           , on_add_to_important_dir_clicked
                                           , misc_toolbar)
add_to_important_dirs_button.setMinimumHeight(height_of_toolbar_buttons)
QToolTip.add(add_to_important_dirs_button
             , _("Add the directory that is currently visible in the left panel to the important directories."))

#QObject.connect(add_to_important_dirs_button, SIGNAL("clicked()")
#                , on_add_to_important_dir_clicked)

#
def on_remove_from_important_dir_clicked():
    global important_directories
    if cur_dir in important_directories:
        if cur_dir == "/":
            QMessageBox.critical(mainVBox, _("Error"),
                                 _("The / directory cannot be removed."))
        else:
            important_directories.remove(cur_dir)
            rebuild_important_directories_menu()
            bookmarksButton.startFlashing(config.gFlashingTimeForButtons,
                                          resetOriginalColorsWhenFinished = True)
            redraw_files() # update icon (folder with heart)
            set_current_dir_button.startFlashing(config.gFlashingTimeForButtons * 3,
                                                 resetOriginalColorsWhenFinished = True)

remove_from_important_dirs_button = QToolButton(QIconSet(QPixmap(config.imgDir + "remove-from-favorites.png"))
                                                , ""
                                                , "unused"
                                                , on_remove_from_important_dir_clicked
                                                , misc_toolbar)

remove_from_important_dirs_button.setMinimumHeight(height_of_toolbar_buttons)
QToolTip.add(remove_from_important_dirs_button
             , _("Remove the directory that is currently visible in the left panel from the important directories."))


#
def depress_all_bulbs():
    global files_whose_light_bulb_is_pressed
    files_whose_light_bulb_is_pressed = []
    redraw_files()
    redraw_arguments()
    redraw_programs()
    listBox['file'].flash_all_items(config.gFlashingTimeForList)
    listBox['argument'].flash_all_items(config.gFlashingTimeForList)
    listBox['program'].flash_all_items(config.gFlashingTimeForList)
    #program_label.startFlashing(config.gFlashingTimeForText
    #                            , resetOriginalColorsWhenFinished = True)
    #argument_label.startFlashing(config.gFlashingTimeForText
    #                             , resetOriginalColorsWhenFinished = True)


remove_all_narrowing_button = FlashingToolButton(QIconSet(QPixmap(config.imgDir
                                                                  + "bulb-delete.png"))
                                                 , ""
                                                 , "unused"
                                                 , depress_all_bulbs
                                                 , misc_toolbar)
remove_all_narrowing_button.setMinimumHeight(height_of_toolbar_buttons)
QToolTip.add(remove_all_narrowing_button,
_("Depresses all pressed file light bulbs, from any directory, \nthus unhiding programs that were previously hidden."))
#QObject.connect(remove_all_narrowing_button
#                , SIGNAL("clicked()"), depress_all_bulbs)






#
def on_purge_clicked():
    global usage_count_of_dir, arguments_of_program, arguments_used_with_any_program
    global last_usage_time_of_dir, sorting_criterion_for_dir
    for x in usage_count_of_dir.keys():
        if not os.path.exists(x):
            print "usage_count_of_dir: removing " + x
            del usage_count_of_dir[x]
    assert forall(usage_count_of_dir.keys()
                  , lambda k:  os.path.exists(k))
    for x in last_usage_time_of_dir.keys():
        if not os.path.exists(x):
            print "last_usage_time_of_dir: removing " + x
            del last_usage_time_of_dir[x]
    assert forall(last_usage_time_of_dir.keys()
                  , lambda k:  os.path.exists(k))
    for x in sorting_criterion_for_dir.keys():
        if not os.path.exists(x):
            print "sorting_criterion_for_dir: removing " + x
            del sorting_criterion_for_dir[x]
    assert forall(sorting_criterion_for_dir.keys()
                  , lambda k:  os.path.exists(k))

    arguments_used_with_any_program = [ a for a in arguments_used_with_any_program
                                        if ( a.name[0] != '/'
                                             or
                                             os.path.exists(a.name))]
    for p in arguments_of_program.keys():
        arguments_of_program[p] = [a for a in arguments_of_program[p]
                                   if ( a.name[0] != '/' or os.path.exists(a.name))]
    redraw_arguments()
    listBox['argument'].flash_all_items(config.gFlashingTimeForList)

purge_button = FlashingToolButton(QIconSet(QPixmap(config.imgDir + "stock_toilet-paper.png"))
                                  , ""
                                  , "unused"
                                  , on_purge_clicked
                                  , misc_toolbar)
purge_button.setMinimumHeight(height_of_toolbar_buttons)
purge_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
QToolTip.add(purge_button, _("Purge the database. This improves the speed and decreases the memory usage.\nClick here once a day. Removes nonexistent files from the arguments list."))
    
#QObject.connect(purge_button, SIGNAL("clicked()"), on_purge_clicked)
#

def on_show_help_clicked():
    help_hbox.show()
    show_help_button.hide()
    global help_bar_visible
    help_bar_visible = True

show_help_button = FlashingToolButton(QIconSet(QPixmap(config.imgDir + "gnome-status.png"))
                                      , ""
                                      , "unused"
                                      , on_show_help_clicked
                                      , misc_toolbar)
show_help_button.setMinimumHeight(height_of_toolbar_buttons)
QToolTip.add(show_help_button, _("Restores the help bar."))
if  help_bar_visible:
    show_help_button.hide()
else:
    show_help_button.show()
#QObject.connect(show_help_button, SIGNAL("clicked()"), on_show_help_clicked)

#




command_hbox = QHBox(mainVBox)
command_hbox.setSpacing(spacing)
clear_command_button = QPushButton( command_hbox)
clear_command_button.setPixmap(QPixmap(config.imgDir + "delete-right-big.png"))
clear_command_button.setMinimumSize( int(44.0 * resolution_multiplier)
                                     , int(32.0 * resolution_multiplier))
QToolTip.add(clear_command_button,
             _("Clears the command area (to the right)."))

def on_clear_command_button_pressed():
    global last_clicked_program
    commandEntry.setText("")
    # BEGIN seems necessary, because it resets itself
    #font = kdecore.KGlobalSettings.fixedFont()
    commandEntry.setCurrentFont(fixed_font)
    # END
    last_clicked_program = None
    redraw_arguments()
    listBox['argument'].flash_all_items(config.gFlashingTimeForList)

    # BEGIN We must not do this: it is the job of on_text_changed to
    # refresh (and flash) the file listbox. 
    #listBox['file'].enable_all_items(False)
    #listBox['file'].update()
    #locationButton.setEnabled(False)
    #END
    

QObject.connect(clear_command_button, SIGNAL("clicked()"), on_clear_command_button_pressed)
clear_command_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

commandEntry = FlashingTextEdit(command_hbox)
#commandEntry.setAcceptDrops(False)
commandEntry.setTextFormat(Qt.PlainText)
fixed_font = kdecore.KGlobalSettings.fixedFont()
commandEntry.setCurrentFont(fixed_font)
#commandEntry.setHScrollBarMode(QScrollView.AlwaysOn)
#commandEntry.setWrapPolicy(QTextEdit.AtWordBoundary)
#commandEntry.setWrapPolicy(QTextEdit.AtWordOrDocumentBoundary)
#commandEntry.setWrapPolicy(QTextEdit.Anywhere)
commandEntry.setWordWrap(QTextEdit.NoWrap)
commandEntry.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
commandEntry.setMaximumHeight(height_of_execute_buttons)

def on_return_pressed_on_command():
    # BEGIN delete the last character, the newline
    commandEntry.moveCursor(QTextEdit.MoveBackward, True)
    commandEntry.delChar()
    # END
    execute(use_terminal = True)
QObject.connect(commandEntry, SIGNAL("returnPressed()"), on_return_pressed_on_command)
def on_text_changed():
    global previous_text
    cur_text = unicode(commandEntry.text())
    cur_p = main_program_in_command(cur_text)
    prev_p = main_program_in_command(previous_text)
    if  cur_p != None and prev_p == None:
#        listBox['file'].startFlashing(config.gFlashingTimeForList,
#                                      resetOriginalColorsWhenFinished = True)
        # BEGIN optimization: instead of redraw_files(centering=True),
        # which was too slow, we call:
#         listBox['file'].enable_all_items(True)
#         listBox['file'].update()
#         locationButton.setEnabled(True)
        # END
        help_label.setText(message_you_can_now_select_a_file)
        help_label.stopFlashing(resetColor = True)
        #program_label.stopFlashing(resetColor = True)


        #BEGIN set proper flashing for execute buttons
        terminal = program_runs_in_terminal.get(cur_p, None)
        if terminal == "yes":
            execute_button.stopFlashing()
            execute_in_term_button.startFlashing(config.gFlashingTimeForExecuteButtons,
                                                 resetOriginalColorsWhenFinished = False)
        elif terminal == "no":
            execute_in_term_button.stopFlashing()
            execute_button.startFlashing(config.gFlashingTimeForExecuteButtons,
                                         resetOriginalColorsWhenFinished = False)
        else:
            execute_in_term_button.stopFlashing()
            execute_button.stopFlashing()
        # END
    if  cur_p == None and prev_p != None:
        # BEGIN Disable the files that are not executable files.
        #
        # optimization: instead of redraw_files(centering=True), which
        # was too slow, we call:
#         cur_dir_slash = addSlashIfNeeded(cur_dir)
#         files_with_path = [cur_dir_slash+x for x in os.listdir(cur_dir)]
#         executable_files = [o for o in files_with_path if os.path.isfile(o)
#                             if os.access(o, os.X_OK)]
#         listBox['file'].enable_these_items(executable_files)
#         listBox['file'].update()
#         locationButton.setEnabled(False)
        # END
        help_label.setText(initial_message)
        help_label.startFlashing(config.gFlashingTimeForHelpMessage
                                 , resetOriginalColorsWhenFinished = True)
        #program_label.startFlashing(config.gFlashingTimeForHelpMessage
         #                           , resetOriginalColorsWhenFinished = True)

        execute_button.stopFlashing()
        execute_in_term_button.stopFlashing()
    previous_text = cur_text
    
QObject.connect(commandEntry, SIGNAL("textChanged()"), on_text_changed)
commandEntry.show()
commandEntry.setFocus()

QToolTip.add(commandEntry,
             
_("This is the command area. The programs, files and arguments \nyou click will appear here. When you have composed a meaningful\ncommand, you can execute it via the buttons to the right.\nYou can also type text directly, but this is only useful \nfor typing short program names, http locations, etc. \nYou should never type the same string more than once."))



previous_text = ""


exec_in_term_pix = QPixmap(config.imgDir + "gnome-terminal.png")
execute_in_term_button = FlashingPushButton(QIconSet(exec_in_term_pix),
                                        _("Execute"),
                                        command_hbox)
execute_in_term_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
execute_in_term_button.show()
execute_in_term_button.setMinimumHeight(exec_in_term_pix.height() + 5)
QToolTip.add(execute_in_term_button,

_("Open a terminal and execute the command which appears in the command area (to the left)."))

exec_pix = QPixmap(config.imgDir + "action.png")
execute_button = FlashingPushButton(QIconSet(exec_pix),
                                "",
                                command_hbox)
execute_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
execute_button.show()
execute_button.setMinimumHeight(exec_pix.height() + 5)
QToolTip.add(execute_button,
             _("Execute the command, disacarding its output. You won't be able to know\nwhether the program has completed, and whether it was successful or not."))

def main_program_in_command(cmd):
    if len(cmd)>0:
        l = split_spaces_but_not_backslash_space(cmd)
        if len(l)>0:
            p = string.strip(l[0])
            return p
    return None

            
def on_execute_button_pressed():
    execute(use_terminal = False)

def on_execute_in_terminal_button_pressed():
    execute(use_terminal = True)

def execute(use_terminal):
    global last_usage_time_of_program, usage_count_of_program, last_clicked_program
    global arguments_of_program, last_usage_time_of_dir, usage_count_of_dir
    global program_runs_in_terminal, files_whose_light_bulb_is_pressed
    global arguments_used_with_any_program
    
    cmd = unicode(commandEntry.text())
    if len(cmd)>0:
        l = split_spaces_but_not_backslash_space(cmd)
        if len(l)>0:
            p = string.strip(l[0]) # l[0] is the program, and can be a file with path too.
            if ( p in programs
                 or (os.path.isfile(p)
                     and os.access(p, os.X_OK))):
                print "execute: inside"
                # BEGIN increase stats for files
                files = [f for f in l if os.path.exists(f)]
                for f in files:
                    usage_count_of_dir[f] = usage_count_of_dir.get(f,0) +1
                    last_usage_time_of_dir[f] = datetime.datetime.now()
                # END
                arguments = l[1:]
                # BEGIN set statistics concerning program p
                usage_count_of_program[p] = usage_count_of_program.get(p, 0) +1
                last_usage_time_of_program[p] = datetime.datetime.now()

                rebuild_important_directories_menu()
                
                if arguments_of_program.get(p, None) == None:
                    arguments_of_program[p] = []
                for a in arguments:
                    fi = find(arguments_of_program[p],
                              lambda pair: pair.name == a)
                    if fi ==None:
                        arguments_of_program[p].append(
                            argument_info(a,1, datetime.datetime.now()))
                    else:
                        arguments_of_program[p].remove( fi)
                        arguments_of_program[p].append(
                            argument_info(a, fi.usage_count +1,
                                          datetime.datetime.now()))
                if use_terminal:
                    program_runs_in_terminal[p] = "yes"
                else:
                    program_runs_in_terminal[p] = "no"

                if  types_used_with_program.get(p, None) == None:
                    types_used_with_program[p] = Set() # the empty set
                for f in files:
                    if os.path.isdir(f):
                        types_used_with_program[p].add("<directory>")
                    else:
                        (type, encoding) = mimetypes.guess_type(f)
                        if type !=None:
                            types_used_with_program[p].add(type)
                        else:
                            types_used_with_program[p].add(ext(f))
                # END

                # BEGIN remember arguments used with ANY program (not
                # p)
                for a in arguments:
                    fi = find(arguments_used_with_any_program,
                              lambda pair: pair.name == a)
                    if fi ==None:
                        arguments_used_with_any_program.append(
                            argument_info(a,1, datetime.datetime.now()))
                    else:
                        arguments_used_with_any_program.remove( fi)
                        arguments_used_with_any_program.append(
                            argument_info(a, fi.usage_count +1,
                                          datetime.datetime.now()))

                # END
                filterQuickText['program'].setText("")
                commandEntry.setText("")
                commandEntry.setFamily("mono") # seems necessary, because it resets itself
                last_clicked_program = None
                cmd = kdesu_start + cmd + kdesu_end
                print cmd
                if use_terminal:
                    def replace_file_with_cross(f):
                        if os.path.exists(f):
                            return "X"
                        else:
                            return f
                    args2 = [replace_file_with_cross(a) for a in arguments]
                    title = p +  reduce(lambda x, y: x + "  " + y
                                        , args2
                                        , "")
                    #os.system( "\"" + config.parser_file_dir + "/one-finger-terminal.py\" -T \""
                    #       + title + "\" --command \" "+ cmd +  "\" --pixmap \""
                    #       + config.imgDir + "onefinger.png\" & " )
                    os.system( "konsole --noclose  -T \"" + title + "\" -e " +  cmd +  " & " )
                else:
                    os.system(cmd + " &")
                    #os.system(config.parser_file_dir + "/one-finger-terminal.py &")
                files_whose_light_bulb_is_pressed = []
                redraw_programs()
                redraw_files()
                redraw_arguments()
                if run_as_root_button.isOn():
                    run_as_root_button.toggle()

            else:
                QMessageBox.critical(mainVBox
                                  , "Program not recognized"
                                  , "The program \"" + p + "\" has been rejected because it was not in the path, \nit was not an executable file and it was not a known built-in command.")


QObject.connect(execute_button, SIGNAL("clicked()"), on_execute_button_pressed)
QObject.connect(execute_in_term_button, SIGNAL("clicked()"), on_execute_in_terminal_button_pressed)


run_as_root_button = QPushButton(QIconSet(QPixmap(config.imgDir + "hidden.png")),
                                 _(""),
                                 command_hbox)
run_as_root_button.setToggleButton(True)
run_as_root_button.show()
run_as_root_button.setMinimumHeight(exec_pix.height() + 5)
run_as_root_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(run_as_root_button,
             _("Execute the command as root (i.e. with administrator privileges). After pushing this button,\nyou have to push one of the two buttons to the left."))
kdesu_start = ""
kdesu_end = ""
def on_run_as_root_toggled(on):
    print "on_run_as_root_toggled"
    global kdesu_end, kdesu_start
    if on:
        kdesu_start = "kdesu -c \""
        kdesu_end = "\""
    else:
        kdesu_end = ""
        kdesu_start = ""
QObject.connect(run_as_root_button, SIGNAL("toggled(bool)"), on_run_as_root_toggled)


drag_to_button = QPushButton(QIconSet(QPixmap(config.imgDir + "stock_drag-mode.png")),
                             _("Drag to..."),
                             command_hbox)
drag_to_button.show()
drag_to_button.setMinimumHeight(exec_pix.height() + 5)
drag_to_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
def uris_in_command_area():
     cur_text = unicode(commandEntry.text())
     l = [string.strip(x)
          for x in split_spaces_but_not_backslash_space(cur_text)]
     def is_uri_in_unicode_uri_format(f):
         r = string.find(f, "file:/")
         r2 = string.find(f, "http:/")
         r3 = string.find(f, "ftp:/")
         return r == 0 or r2 == 0 or r3 ==0
     return [f for f in l
             if ( os.path.exists(f)
                  or is_uri_in_unicode_uri_format(f))]
def on_drag_to_clicked():
    if uris_in_command_area() == []:
        QMessageBox.critical(mainVBox,
                             _("Dragging Error - No file in command area")
                             ,_("The button you have pushed is for dragging files (or URIs) from OneFinger\nto another window. First, click some files so that they appear in the command\narea. Then, press the button."))
    else:
        global dragging_state
        dragging_state = 'waiting-for-window'
        cancel_drag_button.show()
        windowlist_window.move(QPoint(windowlist_x, windowlist_y))
        windowlist_window.show()
        kdecore.KWin.forceActiveWindow(windowlist_window.winId())
        # BEGIN hack to remove the highlighted item @@
        listBox['window'].lastMousePos = None
        listBox['window'].make_changes_visible() 
        # END
        windowlist_help_label.setText(_("Click the window you want to drag the files to."))
        windowlist_help_label.startFlashing(config.gFlashingTimeForHelpMessage
                                            , resetOriginalColorsWhenFinished = True)
        
QObject.connect(drag_to_button, SIGNAL("clicked()"), on_drag_to_clicked)
        
QToolTip.add(drag_to_button,
             
 _("This button is for dragging files (or URIs) from OneFinger to\nanother window. First, click some files so that they appear in the\ncommand area (to the left). Then, press this button. You'll be asked\nfor the target window, which will be automatically activated."))

initial_message =  _("To begin a new command, click a program or a file's light-bulb icon.")
help_hbox = QHBox(mainVBox)
help_hbox.setMargin(spacing)
help_hbox.setPaletteBackgroundColor(help_hbox.palette().active().mid())
help_hbox.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
help_hbox.setSpacing(spacing)
if help_bar_visible:
    help_hbox.show()
else:
    help_hbox.hide()
help_label = FlashingLabel( initial_message, help_hbox)
help_label.setAlignment(Qt.AlignRight |Qt.AlignVCenter)
help_label.startFlashing(config.gFlashingTimeForHelpMessage
                         , resetOriginalColorsWhenFinished = True)
message_explaining_programs_and_narrowing = _("The \"programs\" list below contains the programs currently installed\nin your \"path\". You must always click a program first, to begin a\ncommand; only then you will be able to click files and other\narguments. However, if you want to open a file which you already see,\nand don't want to click a program first, because the program list is\ntoo long, you can click on the \"light bulb\" icon besides the file:\nthis will suggest the programs you can use to view that file, thus \nallowing a document-based interaction.")
QToolTip.add(help_label, message_explaining_programs_and_narrowing)
more_help_button = QPushButton(QIconSet(QPixmap(config.imgDir + "gnome-status.png"))
                               ,_("More details"), help_hbox)
more_help_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(more_help_button, _("Gives you more informations."))
def on_more_details_button_clicked():
    QMessageBox.information(mainVBox, _("More details")

, _("Read the tooltips for more details. Tooltips are messages that appear\nwhen you place the mouse over widgets and wait some time without\nmoving it.\n\nFor example, close this window, then place the mouse over\nthe flashing message. You'll have additional details.")
    )
QObject.connect(more_help_button, SIGNAL("clicked()")
                , on_more_details_button_clicked)

hide_help_button=QPushButton(QIconSet(QPixmap(config.imgDir + "delete.png"))
                             , _("Hide")
                             , help_hbox)
hide_help_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(hide_help_button, _("Hides this help bar to gain more space. You will be able to activate it again \nwith a toolbar button."))
def on_hide_help():
    help_hbox.hide()
    global help_bar_visible
    help_bar_visible = False
    show_help_button.show()
    show_help_button.startFlashing(config.gFlashingTimeForButtons *4
                                   , resetOriginalColorsWhenFinished = True)
QObject.connect(hide_help_button, SIGNAL("clicked()")
                , on_hide_help)
mainGrid = QGrid(3, mainVBox)
mainGrid.setSpacing(spacing)
mainGrid.setMargin(spacing)
mainGrid.show()

hBoxSpacing = 3

# FIRST ROW
# create three hBox, indexed by panel
for k in ['file','program',  'argument']:
    hBox[k] = QHBox(mainGrid)
    hBox[k].show()
    hBox[k].setSpacing(hBoxSpacing)

    if k != 'file':
        icon[k] = QLabel(hBox[k])
        icon[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        icon[k].show()

#if not window_panel_visible:
#    hBox['window'].hide()
set_current_dir_button = FlashingPushButton(QIconSet(file_to_icon(cur_dir))
                                        , ""
                                        , hBox['file'])
set_current_dir_button.show()
set_current_dir_button.setToggleButton(True)
def on_current_dir_changed():
    os.chdir(cur_dir)
QObject.connect(set_current_dir_button, SIGNAL("clicked()"), on_current_dir_changed)
set_current_dir_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(set_current_dir_button,

_("Set this directory as the current bash \"working directory\". This is needed\nbecause some programs such as cvs behave in a different way according to\nthe current directory. Using this button is like typing \"cd <dir>\" before \nexecuting a command."))

locationButton = WrappingButton('text will change', hBox['file'], data = None, toggle = False)
QObject.connect(locationButton, PYSIGNAL("pressed"), onLocationButtonPressed)
locationButton.show()

QToolTip.add(locationButton,

_("This button contains the name of the directory that is currently\nvisible in the file panel (below)."))


narrow_dir_button = QPushButton(QIconSet(QPixmap(config.imgDir + "bulb.png")),
                                "",
                                hBox['file'])
narrow_dir_button.setToggleButton(True)
narrow_dir_button.show()
narrow_dir_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
narrow_dir_button.setMinimumWidth( int(config.width_of_scroll_buttons * resolution_multiplier))

# This is called also when I call setOn via code :-( Therefore I have
# to be careful not to insert duplicates or remove when the dir isn't
# there. See (1) and (2)
def on_narrow_dir_button_toggled(on):
    global files_whose_light_bulb_is_pressed
    if on:
        remove_all_narrowing_button.startFlashing(config.gFlashingTimeForButtons
                                                  , resetOriginalColorsWhenFinished = True)
        extens_before = types_selected_for_prog_narrowing()
        if not cur_dir in files_whose_light_bulb_is_pressed: # (1)
            files_whose_light_bulb_is_pressed.append(cur_dir)
        # BEGIN we have to synchronize the light bulbs of the other panel :-(
        listBox['argument'].set_pushbuttons_of_this_item_if_it_exists(
            pressed = True , item_data = cur_dir)
        listBox['argument'].make_changes_visible()
        # END        extens_after = types_selected_for_prog_narrowing()
        extens_after = types_selected_for_prog_narrowing()
        if len(extens_after) > len(extens_before):
            redraw_programs()
            listBox['program'].flash_all_items(config.gFlashingTimeForList)
            #program_label.startFlashing(config.gFlashingTimeForText
            #                            , resetOriginalColorsWhenFinished = True)

    else:
        extens_before = types_selected_for_prog_narrowing()
        if cur_dir in files_whose_light_bulb_is_pressed: # (2)
            files_whose_light_bulb_is_pressed.remove(cur_dir)
        # BEGIN we have to synchronize the light bulbs of the other panel :-(
        listBox['argument'].set_pushbuttons_of_this_item_if_it_exists(
            pressed = False , item_data = cur_dir)
        listBox['argument'].make_changes_visible()
        # END        extens_after = types_selected_for_prog_narrowing()
        extens_after = types_selected_for_prog_narrowing()
        if len(extens_after) < len(extens_before):
            redraw_programs()
            #program_label.startFlashing(config.gFlashingTimeForText
            #                            , resetOriginalColorsWhenFinished = True)
            listBox['program'].flash_all_items(config.gFlashingTimeForList)
QObject.connect(narrow_dir_button, SIGNAL("toggled(bool)"), on_narrow_dir_button_toggled)

QToolTip.add(narrow_dir_button,

_("Light bulbs like this one can be found beside each file. Clicking a\nlight bulb means \"Suggest me a program to open this file\". At the\nbeginning, OneFinger will not have much to suggest but, as time goes\nby, OneFinger will learn which programs can open which files, by\nwatching the way you work."))

program_label = FlashingLabel(_('Programs'), hBox['program'])
program_label.show()
#program_label.startFlashing(config.gFlashingTimeForHelpMessage
 #                           , resetOriginalColorsWhenFinished = True)
icon['program'].setPixmap(QPixmap( config.imgDir + "person.png"))
QToolTip.add(program_label, message_explaining_programs_and_narrowing)

show_arguments_with_any_program_button = QPushButton(QIconSet(QPixmap(config.imgDir + "glasses-crossed.png")),
                                "",
                                hBox['program'])
show_arguments_with_any_program_button.show()
show_arguments_with_any_program_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
show_arguments_with_any_program_button.setMinimumWidth( int(config.width_of_scroll_buttons * resolution_multiplier))
QToolTip.add(show_arguments_with_any_program_button

, _("Show the arguments used with any program. This button reverses\nthe effect of the other \"glasses-shaped\" buttons, which narrow the\narguments to only those ones used with a given program."))

def on_show_args_with_any_program_clicked():
    global last_clicked_program
    last_clicked_program = None
    listBox['argument'].flash_all_items(config.gFlashingTimeForList)
    redraw_arguments()
QObject.connect(show_arguments_with_any_program_button, SIGNAL("clicked()")
                , on_show_args_with_any_program_clicked)
    
argument_label = FlashingLabel(_('Arguments'), hBox['argument'])
argument_label.show()
icon['argument'].setPixmap(QPixmap( config.imgDir + "options.png"))

QToolTip.add(argument_label,
        
_("The list below contains the \"arguments\" you passed to the active\nprogram in your previous commands. Arguments can be files, or options\nsuch as \"--human-readable\". This panel exists mainly because it's\ndifficult to remember some options you use very rarely, like \n\"update-alternatives --config x-session-manager\". This way, you\ndon't have to read the man page each time you want to use a rarely use\ncommand."))


#
show_files_in_arg_panel_button = QPushButton(QIconSet(QPixmap(config.imgDir + "file-manager.png")),
                                               "",
                                               hBox['argument'])
show_files_in_arg_panel_button.show()
show_files_in_arg_panel_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
show_files_in_arg_panel_button.setMinimumWidth( int(config.width_of_scroll_buttons * resolution_multiplier))
show_files_in_arg_panel_button.setToggleButton(True)
QToolTip.add(show_files_in_arg_panel_button

, _("If pressed, files are shown in the argument panel. Otherwise, only options such as --long are shown.\nLeave this pressed, unless you can't find the option you need."))
show_files_in_arg_panel = True
show_files_in_arg_panel_button.setOn(True)
def on_show_files_in_arg_panel_button_toggled(on):
    global show_files_in_arg_panel
    show_files_in_arg_panel = not show_files_in_arg_panel
    listBox['argument'].flash_all_items(config.gFlashingTimeForList)
    redraw_arguments()
QObject.connect(show_files_in_arg_panel_button, SIGNAL("toggled(bool)")
                , on_show_files_in_arg_panel_button_toggled)


#
# window_label = FlashingLabel(_('Windows'), hBox['window'])
# window_label.show()
# icon['window'].setPixmap(QPixmap( config.imgDir + "window.png"))

# QToolTip.add(window_label,
        
# _("The list below contains the open windows. Clicking on a window name\nactivates that window, which makes OneFinger act like a taskbar. In\naddition, beside each window there is a \"drag\" button, which can be\nused to drag some files onto that window. This button is useful\nbecause it automatically focuses the target window, so you don't need\nto resize either OneFinger or the target window. To use the dragging\nbutton: first, click the files you want to send, so that they appear\nin the command area; then, press the drag button beside the\nwindow you want to send the files to. The window will appear, covering\nOneFinger. Finally, press the button to the right spot on the target\nwindow.")

# )



# close_window_panel_button = QPushButton(QIconSet(QPixmap(config.imgDir + "delete.png")),
#                                 "",
#                                 hBox['window'])
# close_window_panel_button.show()
# close_window_panel_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
# close_window_panel_button.setMinimumWidth( int (config.width_of_scroll_buttons * resolution_multiplier))
# QToolTip.add(close_window_panel_button

# , _("Close the window panel. Beware! You will loose the capability\nto drag from OneFinger to external applications and to use\nOneFinger as a taskbar."))

# def on_close_window_panel_clicked():
#     global window_panel_visible
#     window_panel_visible = False
#     hBox['window'].hide()
#     listBox['window'].hide()
#     sortBox['window'].hide()
#     filterBox['window'].hide()
#     show_window_panel_button.show()
#     show_window_panel_button.startFlashing(config.gFlashingTimeForButtons *4
#                                            , resetOriginalColorsWhenFinished = True)

# QObject.connect(close_window_panel_button, SIGNAL("clicked()")
#                 , on_close_window_panel_clicked)
    


# SECOND ROW
def on_list_item_togglebutton_pressed(list, item, button_meaning):
    if button_meaning == "narrow":
        global files_whose_light_bulb_is_pressed
        extens_before = types_selected_for_prog_narrowing()
        files_whose_light_bulb_is_pressed.append(item.data)
        # BEGIN we have to synchronize the light bulbs of the other panel :-(
        k = find([ 'file', 'program', 'argument'], lambda x: listBox[x] is list)
        if k == 'file':
            listBox['argument'].set_pushbuttons_of_this_item_if_it_exists(
                pressed = True , item_data = item.data)
            listBox['argument'].make_changes_visible()
            listBox['file'].make_changes_visible() # update the bulbs
        else:
            listBox['file'].set_pushbuttons_of_this_item_if_it_exists(
                pressed = True , item_data = item.data)
            listBox['file'].make_changes_visible()
            listBox['argument'].make_changes_visible() # update the bulbs
            if cur_dir == item.data:
                print "setting on the icon!"
                narrow_dir_button.setOn(True)
        # END
        extens_after = types_selected_for_prog_narrowing()
        if len(extens_after) > len(extens_before):
            redraw_programs()
            listBox['program'].flash_all_items(config.gFlashingTimeForList)
        remove_all_narrowing_button.startFlashing(config.gFlashingTimeForButtons,
                                                  resetOriginalColorsWhenFinished = True)
    else:
        raise "by god"
    
def on_list_item_togglebutton_released(list, item, button_meaning):
    if button_meaning == "narrow":
        k = find([ 'file', 'program', 'argument'], lambda x: listBox[x] is list)
        global files_whose_light_bulb_is_pressed
        extens_before = types_selected_for_prog_narrowing()
        files_whose_light_bulb_is_pressed.remove(item.data)
        # BEGIN we have to synchronize the light bulbs of the other panel :-(
        if k == 'file':
            listBox['argument'].set_pushbuttons_of_this_item_if_it_exists(
                pressed = False , item_data = item.data)
            listBox['argument'].make_changes_visible()
            listBox['file'].make_changes_visible() # update the bulb
        else:
            listBox['file'].set_pushbuttons_of_this_item_if_it_exists(
                pressed = False , item_data = item.data)
            listBox['file'].make_changes_visible()
            listBox['argument'].make_changes_visible() # update the bulb
            if cur_dir == item.data:
                print "setting off the icon!"
                narrow_dir_button.setOn(False)
        # END
        extens_after = types_selected_for_prog_narrowing()
        if len(extens_after) < len(extens_before):
            redraw_programs()
            listBox['program'].flash_all_items(config.gFlashingTimeForList)

    else:
        raise "by god"

# def on_list_item_button_released(list, item, meaning):
#     k = find([ 'file', 'program', 'argument', 'window'], lambda x: listBox[x] is list)
#     if k == 'window':
#         print "window"
#         print ("item = " + str(item) + " , item_whose_button_was_last_pressed = "
#                + str (item_whose_button_was_last_pressed))
#         if item.data == item_whose_button_was_last_pressed:
#             assert meaning == 'drag'
#             cur_text = unicode(commandEntry.text())
#             l = [string.strip(x)
#                  for x in split_spaces_but_not_backslash_space(cur_text)]
#             def is_uri_in_unicode_uri_format(f):
#                 r = string.find(f, "file:/")
#                 r2 = string.find(f, "http:/")
#                 r3 = string.find(f, "ftp:/")
#                 return r == 0 or r2 == 0 or r3 ==0
#             files = [f for f in l
#                      if ( os.path.exists(f)
#                           or is_uri_in_unicode_uri_format(f))]
#             if files == []:
#                 QMessageBox.critical(mainVBox,
#                                      _("Dragging Error - No file in command area")
#                                      ,_("The button you have pushed is for dragging files (or URIs) from OneFinger\nto another window. First, click some files so that they appear in the command\narea. Then, press the button."))
         
#             else:
#                 kdecore.KWin.activateWindow(item.data)
#                 d = QUriDrag(files, mainWindow)
#                 d.dragCopy()

def on_list_item_button_pressed(list, item, button_meaning):
    global item_whose_button_was_last_pressed
    item_whose_button_was_last_pressed = item.data # remember for when
                                                   # it's released
    k = find([ 'file', 'program', 'argument', 'window'], lambda x: listBox[x] is list)
    if k == 'file' or k=='argument':
        if button_meaning == "enter directory":
            global recent_dirs, cur_dir
            recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
            cur_dir = item.data
            filterQuickText[k].clear()
            usage_count_of_dir[cur_dir] = usage_count_of_dir.get(cur_dir, 0) + 1
            last_usage_time_of_dir[cur_dir] = datetime.datetime.now()
            recent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                            resetOriginalColorsWhenFinished = True)
            rebuild_recent_locations_menu(recent_dirs)
            rebuild_parent_dirs_menu()
            # BEGIN don't change; getSortingCriterion is an english string:
            cr = _(sorting_criterion_for_dir.get(cur_dir, "Recent usage"))
            # END
            sortBox[k].setCurrentText(cr)
            redraw_files()
            recent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                            resetOriginalColorsWhenFinished = True)
            parent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                            resetOriginalColorsWhenFinished = True)
            if k == 'argument':
                listBox['file'].flash_all_items(config.gFlashingTimeForList)

        else:
            raise "by god"
    
    elif k == 'program':
        global last_clicked_program
        last_clicked_program = item.data
        redraw_arguments()
        listBox['argument'].flash_all_items(config.gFlashingTimeForList)
    elif k == 'window':
        print ' pass '
        pass # wait until the button is released
    else:
        raise "goddamn it"


def make_string_bash_compatible(f):
    t = string.replace(f, '\"', '\"')
    return '\"' + f + '\"'


previously_activated_window = None
def on_list_item_pressed(box, item):
    def is_utility_window(w):
        inf = kdecore.KWin.windowInfo(w, kdecore.NET.WMWindowType)
        t =  inf.windowType(kdecore.NET.NormalMask
                            | kdecore.NET.UtilityMask
                            | kdecore.NET.DesktopMask
                            | kdecore.NET.DockMask
                            | kdecore.NET.ToolbarMask)
        return t == kdecore.NET.Utility

    k = find([ 'file', 'program','argument', 'window'], lambda x: listBox[x] is box)
    if k != 'window':
        commandEntry.startFlashing(3, True)

    if k == 'program':
        commandEntry.insert(item.data + "  ") # fires on_text_changed
        global last_clicked_program
        last_clicked_program = item.data
        # BEGIN DON'T DO THIS! It is the job of on_text_changed to
        # flash the file listbox; doing this here would only result in
        # duplicated redrawing of the file list.
        #
        #if the_string_specifies_a_program(...) == False:
        #listBox['file'].startFlashing(config.gFlashingTimeForList,
        #                             resetOriginalColorsWhenFinished = True)
        #redraw_files(centering = True)
        #
        #END
        listBox['argument'].flash_all_items(config.gFlashingTimeForList)
        redraw_arguments()
    elif k == 'argument':
        t = make_string_bash_compatible(item.data)
        commandEntry.insert(t + "  ")
    elif k == 'file': # files are quoted
        t = make_string_bash_compatible(item.data)
        commandEntry.insert(t + "  ")
    elif k == 'window':
        if dragging_state == 'waiting-for-window':
            pass # wait for button release
        elif dragging_state == 'no-drag':
            win_pid = pid(item.data)
            util_windows_to_bring_on_top = [q for q in windows_sorted_by_creation_date
                                            if pid(q) == win_pid
                                            if is_utility_window(q)]

            for x in util_windows_to_bring_on_top:
                kdecore.KWin.activateWindow(x)
            kdecore.KWin.activateWindow(item.data)
            windowlist_window.hide()
            
        
    commandEntry.setFocus()


def on_list_item_released(box, item):
    global dragging_state
    if dragging_state == 'waiting-for-window':
        dragging_state = 'no-drag'
        cancel_drag_button.hide()
        kdecore.KWin.activateWindow(item.data)
        d = QUriDrag(uris_in_command_area(), mainWindow)
        d.dragCopy() 
        windowlist_window.hide()
        windowlist_help_label.setText(windowlist_default_message)


for k in [ "file", "program", "argument"]:
    listBox[k] = MyListBox(frameVisible = False, parent = mainGrid
                           , resolution_mult = resolution_multiplier)
listBox['window'] = MyListBox(frameVisible = False, parent = windowlist_vbox
                              , resolution_mult = resolution_multiplier)
for k in [ "file", "program", "argument", 'window']:
    listBox[k].show()
    QObject.connect(listBox[k], PYSIGNAL("item_button_pressed"), on_list_item_button_pressed)
    #    QObject.connect(listBox[k], PYSIGNAL("item_button_released"), on_list_item_button_released)
    QObject.connect(listBox[k], PYSIGNAL("item_togglebutton_pressed"), on_list_item_togglebutton_pressed)
    QObject.connect(listBox[k], PYSIGNAL("item_togglebutton_released"), on_list_item_togglebutton_released)
    QObject.connect(listBox[k], PYSIGNAL("item_pressed"), on_list_item_pressed)
    QObject.connect(listBox[k], PYSIGNAL("item_released"), on_list_item_released)



# THIRD ROW
for k in [ "file", "program","argument", 'window']:
    if k != 'window':
        filterBox[k] = QHBox(mainGrid)
    else:
        filterBox[k] = QHBox(windowlist_vbox)
    filterBox[k].show()
    filterBox[k].setSpacing(3) # @@ TODO: scale?

    filterClearButton[k] = QPushButton(filterBox[k])
    filterClearButton[k].show()
    filterClearButton[k].setIconSet(QIconSet(QPixmap(config.imgDir + 'delete-right.png')))
    filterClearButton[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    filterQuickText[k] = MyLineEdit(filterBox[k])
    filterQuickText[k].show()
    QObject.connect(filterQuickText[k], PYSIGNAL("timer"), onFilterKeyTimer)
    QObject.connect(filterQuickText[k], SIGNAL("returnPressed()"), onEnterPressedOnFilter)

    filterButton[k] = QPushButton(QIconSet(QPixmap(config.imgDir + "mbuto.png")),
                                  _("Recent filters"), filterBox[k])
    filterButton[k].show()

    filterMenu[k] = MyPopupMenu(filterButton[k])
    QObject.connect(filterMenu[k], PYSIGNAL("itemActivated"), onRecentFilterClicked)
    filterButton[k].setPopup(filterMenu[k])


QToolTip.add(filterClearButton['file'],
 _("This button clears the filter area to the right. This means you will see all the \nfiles that were previously hidden."))
QToolTip.add(filterClearButton['program'],
 _("This button clears the filter area to the right. This means you will see all the \nprograms that were previously hidden."))
QToolTip.add(filterClearButton['argument'],
 _("This button clears the filter area to the right. This means you will see all the \narguments that were previously hidden."))

QToolTip.add(filterQuickText['file'],
 _("This is the file filter area.\nIf you type some text here, only those files whose name \ncontains the text will be shown."))
QToolTip.add(filterQuickText['program'],
 _("This is the program filter area.\nIf you type some text here, only those programs whose name \ncontains the text will be shown."))
QToolTip.add(filterQuickText['argument'],
 _("This is the argument filter area.\nIf you type some text here, only those arguments whose name \ncontains the text will be shown."))


QToolTip.add(filterButton['file'],
 _("Use this button to reapply a previous filter to the files list. \nFile filters typically need to be applied again and again: \nthis way you can type them only once and use them forever."))

QToolTip.add(filterButton['program'],
 _("Use this button to reapply a previous filter to the programs list."))

QToolTip.add(filterButton['argument'],
 _("You can use this button to reapply a previous filter to the arguments list."))

rebuildFilterMenus()
QObject.connect(filterClearButton['file'], SIGNAL("clicked()"), clearFileFiltersAndUpdate)
QObject.connect(filterClearButton['program'], SIGNAL("clicked()"), clearProgramFiltersAndUpdate)
QObject.connect(filterClearButton['argument'], SIGNAL("clicked()"), clearArgumentFiltersAndUpdate)
QObject.connect(filterClearButton['window'], SIGNAL("clicked()"), clearWindowFiltersAndUpdate)

# FOURTH ROW

kwin_module = kdecore.KWinModule()

windows_sorted_by_creation_date = [x for x in kwin_module.windows()]
windows_sorted_by_usage = [ x for x in kwin_module.windows()] # initially they are the same
windows_sorted_by_usage.reverse()
def on_active_window_changed(w):
    global windows_sorted_by_usage
    # this is called not only when a window is made active, but
    # also when a window is closed! Only after, onWindowRemoved
    # is called! So I must not call rebuild_window_menu if the
    # window has no wid. BUG FIXED.
    if kwin_module.hasWId(w):
        windows_sorted_by_usage.remove(w)
        windows_sorted_by_usage = [w] + windows_sorted_by_usage
        tx = sortBox['window'].currentText()
        recent_usage = tx == _("Recent usage")
        redraw_windows(remember_scroll = not recent_usage)
    else:
        pass
        # do nothing. on_window_removed will be called shortly, and
        # will remove the WId from the list

try:
    QObject.connect(kwin_module
                    , SIGNAL("activeWindowChanged(WId)"), on_active_window_changed)



except RuntimeError:
    pykde_supports_wid = False
    error_msg = _("Since you are running a PyKDE version prior to 3.11.3, you\nwon't be able to sort the windows by recent usage. This is a\nsubstantial limitation, so please consider upgrading your \nPyKDE.")

    print error_msg
    QMessageBox.critical(mainVBox, _("OneFinger - Non critical error"), error_msg)

    # BEGIN fix the PyKDE bug with a polling
    window_timer = QTimer(mainVBox)
    def on_window_timer():
        global windows_sorted_by_creation_date
        windows_sorted_by_creation_date = kwin_module.windows()
        redraw_windows(remember_scroll = True)
    QObject.connect(window_timer, SIGNAL("timeout()"), on_window_timer)
    window_timer.start(1000)
    # END
else:
    pykde_supports_wid = True
    def on_window_added(w):
        global windows_sorted_by_creation_date, windows_sorted_by_usage
        windows_sorted_by_usage = [w] + windows_sorted_by_usage
        windows_sorted_by_creation_date.append(w)
        tx = sortBox['window'].currentText()
        recent_usage = tx == _("Recent usage")
        redraw_windows(remember_scroll = not recent_usage)
#         if w != mainWindow.winId():
#             listBox['window'].startFlashing(config.gFlashingTimeForList,
#                                         resetOriginalColorsWhenFinished = True)
    QObject.connect(kwin_module
                    , SIGNAL("windowAdded(WId)"), on_window_added)
    def on_window_changed(w, changed_properties):
        # this updates the visible name of the window
        if changed_properties & NET.WMName: # without this, OneFinger freezes for seconds!
            redraw_windows(remember_scroll = True) 
         
    #QObject.connect(kwin_module
    #                , SIGNAL("windowChanged(WId, unsigned long)"), on_window_changed)
    def on_window_removed(w):
        global windows_sorted_by_creation_date, windows_sorted_by_usage
        windows_sorted_by_usage.remove(w)
        windows_sorted_by_creation_date.remove(w)
        redraw_windows(remember_scroll = False) # don't change to not(recent_usage)
#         if w != mainWindow.winId():
#             listBox['window'].startFlashing(config.gFlashingTimeForList,
#                                         resetOriginalColorsWhenFinished = True)

    QObject.connect(kwin_module
                    , SIGNAL("windowRemoved(WId)"), on_window_removed)

def onSortBoxItemActivated(box, item):
    global sorting_criterion_for_dir, sorting_criterion_for_programs, sorting_criterion_for_arguments
    assert item in ["Recent usage", "Frequent usage", "Date", "Name", "Size"] # sic! don't change. Don't add _().

    k = find(['program', 'file', 'argument', 'window'], lambda x: box is sortBox[x])

    if k == 'file' :
        print "setting sorting_criterion_for_dir " + cur_dir + " = " + item
        sorting_criterion_for_dir[cur_dir] = item
        redraw_files()
        listBox['file'].flash_all_items(config.gFlashingTimeForList)

    if k == 'program':
        sorting_criterion_for_programs = item
        redraw_programs()
        listBox['program'].flash_all_items(config.gFlashingTimeForList)
    if k== 'argument':
        sorting_criterion_for_arguments = item
        redraw_arguments()
        listBox['argument'].flash_all_items(config.gFlashingTimeForList)

    if k == 'window':
        redraw_windows(remember_scroll = False)
        listBox['window'].flash_all_items(config.gFlashingTimeForList)


for k in ["file", "program", "argument"]:
    sortBox[k] = MyComboBox(mainGrid)
    sortBox[k].setMinimumHeight(int(35.0 * resolution_multiplier))
    sortBox[k].show()
    sortBox[k].insertItem(config.imgDir + "name.png", text = _("Name"), data = "Name")
    sortBox[k].insertItem(config.imgDir + "favourites-small.png", _("Frequent usage"), "Frequent usage")
    sortBox[k].insertItem(config.imgDir + "stock_timer.png", _("Recent usage"), "Recent usage")
    QObject.connect(sortBox[k], PYSIGNAL("itemActivated"), onSortBoxItemActivated)
sortBox['file'].insertItem(config.imgDir + "size.png", _("Size"), "Size")
sortBox['file'].insertItem(config.imgDir + "clock-small.png", _("Date"), "Date")

sortBox['window'] = MyComboBox(windowlist_vbox)
sortBox['window'].setMinimumHeight(int(35.0 * resolution_multiplier))
sortBox['window'].show()
sortBox['window'].insertItem(config.imgDir + "clock-small.png", _("Creation date"), "Date")
if pykde_supports_wid:
    sortBox['window'].insertItem(config.imgDir + "stock_timer.png", _("Recent usage"), "Recent usage")
sortBox['file'].setCurrentText(_(sorting_criterion_for_dir.get(cur_dir, "Recent usage")))
sortBox['program'].setCurrentText(_(sorting_criterion_for_programs))
sortBox['argument'].setCurrentText(_(sorting_criterion_for_arguments))


QToolTip.add(sortBox['argument'],
 _("You can use this button to change the way arguments are ordered."))
QToolTip.add(sortBox['file'],
             
 _("You can use this button to change the order with which files appear in\nthe above list. Each directory will remember this setting independently. You\nshould almost always leave the default order, \"recent usage\", but\nthere are a few exceptions: for your downloads directory, you should\nset a sorting by date, because this way you can quickly find the\nlatest downloaded file. It can also be useful to sort temporarily by\nsize, if you want to quickly find the largest file in order to burn it\nto CD. The other orderings are useless. In particular, sorting by name\nis useless because it's easier to use the filter box above."))


QToolTip.add(sortBox['program'],
             
_("Change the order with which programs appear in the above list. There\nis a dispute about whether ordering by recent usage is better than\nfrequent usage. Personally I prefer recent usage, because frequent\nfiles are still shown among the first items, but \"temporal\ncoherence\" is exploited as well. The default is frequent usage,\nbecause this is what most people seem to use."))
QToolTip.add(sortBox['argument'],
             _("Change the order with which arguments appear in the above list."))



if not pykde_supports_wid:
    sortBox['window'].setCurrentText(_("Creation date"))
else:
    sortBox['window'].setCurrentText(_("Recent usage"))


#


windowlist_toolbar = QToolBar(windowlist_window) 
windowlist_toolbar.show()

def on_close_terminals_pressed():
    print "on_close_terminals_pressed"
    for w in windows_sorted_by_creation_date:
        def is_probably_terminated(w):
            def strip_parent(s):
                print "strip_parent: input: " + s
                assert len(s)>=2
                l = len(s)
                s2 =  s[1:l-1]
                print "strip_parent: output: " + s2
                return s2
            def all_digits(s):
                return forall(s, lambda c: c in digits)
            inf = kdecore.KWin.windowInfo(w)
            nam = str(inf.visibleName())
            print "nam = " + nam
            list = re.findall( r"<\D*?>" ,  nam)
            return is_substring_of('konsole', nam) and len(list)>0 

        print "--"
        if is_probably_terminated(w):
            print "sending quit signal to window " + str(w)
            os.kill(pid(w), signal.SIGKILL)
            # a better way would be:
            #     ri = kdecore.NETRootInfo( qt_xdisplay(),  kdecore.NET.CloseWindow );
            #     ri.closeWindowRequest( w );
            # but it doesn't work... @@ TODO

            
close_terminals_button = QToolButton(QIconSet(QPixmap(config.imgDir + "gnome-terminal.png"))
                                     , _("Close terminated output windows")
                                     , "unused"
                                     , on_close_terminals_pressed
                                     , windowlist_toolbar)
close_terminals_button.setMinimumHeight(height_of_toolbar_buttons)
close_terminals_button.setUsesTextLabel(True)
close_terminals_button.setTextPosition(QToolButton.BesideIcon)
QToolTip.add(close_terminals_button
             , _("Closes all konsole windows whose associated command is terminated."))


#

redraw_files()
redraw_programs()
redraw_arguments()
redraw_windows(remember_scroll = False)

mainWindow.showMaximized()

if (  windowlist_x_read_from_db == None or windowlist_y_read_from_db == None
      or windowlist_wt_read_from_db == None or windowlist_ht_read_from_db == None):
    windowlist_window.showMaximized()
else:
    windowlist_window.show()
    # BEGIN this must be before move, otherwise the win could be resized
    windowlist_window.resize(QSize(windowlist_wt_read_from_db, windowlist_ht_read_from_db))
    # END
    windowlist_window.move(QPoint(windowlist_x_read_from_db, windowlist_y_read_from_db))

    

# BEGIN Quick hack to set the icon in the window bar, because
# KMainWindow.setIcon() does not work (kde bug?).
kdecore.KWin.setIcons(mainWindow.winId()
                      , QPixmap(config.imgDir + "onefinger.png")
                      , QPixmap( config.imgDir + "onefinger.png"))
kdecore.KWin.setIcons(windowlist_window.winId()
                      , QPixmap(config.imgDir + "onefinger.png")
                      , QPixmap( config.imgDir + "onefinger.png"))
# END

rebuild_recent_locations_menu(recent_dirs)
rebuild_important_directories_menu()
rebuild_parent_dirs_menu()

# BEGIN test file lock
#fd = os.open("/root/testicon/ChangeLog", os.O_EXCL | os.O_RDWR)
#print "fd = " + str(fd)
# END
applic.exec_loop()

save_database()
